# distutils: language = c++
# distutils: libraries = gmp gmpxx pplite m flint

cimport cython

from gmpy2 cimport import_gmpy2, mpz, mpz_t, GMPy_MPZ_From_mpz, MPZ_Check
from libcpp.vector cimport vector as cppvector
from .integer_conversions cimport FLINT_Integer_to_Python, Python_int_to_FLINT_Integer, FLINT_Rational_to_Python, Python_float_to_FLINT_Rational
from .constraint cimport Constraint
from .generators cimport PPliteGenerator
from .linear_algebra cimport Variable, Linear_Expression, Affine_Expression
from .intervals cimport Interval

cdef class NNC_Polyhedron(object):
    r"""
    Wrapper for PPLite's ``Poly`` class.
    
    INPUT:

    - dim_type (int), spec_elem (string "universe" or "empty"), and topology ("closed" or "nnc") xor,

    - nnc_poly, :class:`NNC_Polyhedron` xor,

    - cons - a list of :class:`Constraint`.

    OUTPUT:

    A :class:`NNC_Polyhedron`

    EXAMPLES::

    Construct an empty polyhedron:

    >>> from pplite import *
    >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "empty", topology = "nnc")
    >>> P
    false

    Construct the universe:

    >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
    >>> P

    Define the open first quadrant R^2 by adding constraints:

    >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
    >>> x = Variable(0)
    >>> y = Variable(1)
    >>> P.add_constraint(x>0)
    >>> P.add_constraint(y>0)
    >>> P
    x0>0, x1>0

    We can define a polyhedron with another polyhedron:

    >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
    >>> x = Variable(0)
    >>> y = Variable(1)
    >>> P.add_constraint(x>0)
    >>> P.add_constraint(y>0)
    >>> P
    x0>0, x1>0   
    >>> P_2 = NNC_Polyhedron(nnc_poly = P)
    >>> P_2
    x0>0, x1>0

    We can add generators in the form of points, closure points, lines, and rays to a polyhedron:

    >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
    >>> x = Variable(0)
    >>> y = Variable(1)
    >>> P.add_constraint(x>0)
    >>> P.add_constraint(y>0)
    >>> P_2 = NNC_Polyhedron(nnc_poly = P)
    >>> P_2.add_generator(Point())  # This is a short cut for the origin
    >>> P_2
    x0>=0, x1>=0

    Add several constraints or generators at a time:

    >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
    >>> x = Variable(0)
    >>> y = Variable(1)
    >>> P.add_constraints([x>0, y>0])
    >>> P
    x0>0, x1>0

    Directly define a polyhedron from a list of constraints:

    >>> P_2 = NNC_Polyhedron(cons = [x>0, y>0])
    >>> P == P_2
    True    
    """
    def __init__(self, **kwrds):
        """
        TESTS::
        >>> from pplite import NNC_Polyhedron
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "empty", topology = "nnc")
        >>> P
        false
        >>> P_2 = NNC_Polyhedron(nnc_poly = P)
        >>> P_2
        false
        >>> P_2.equals(P)
        True
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, Constraint
        >>> A = Variable(0)
        >>> P_3 = NNC_Polyhedron(dim_type = 1, spec_elem = "universe", topology = "nnc")
        >>> P_3.add_constraint(A >= 0)
        >>> P_3
        x0>=0
        >>> P_4 = NNC_Polyhedron(nnc_poly = P_3)
        >>> P_4.add_constraint(A >= -10)
        >>> P_4.add_constraint(A >= -5)
        >>> P_4.add_constraint(A >= -2)
        >>> P_4.add_constraint(A >= -1)
        >>> P_4.add_constraint(A >= 0)
        >>> P_3.equals(P_4)
        True
        >>> B = Variable(1)
        >>> cons_list = [A >= 0, B == 5]
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P.add_constraints(cons_list)
        >>> P
        x1-5==0, x0>=0
        """
        cdef dim_type dd
        cdef Spec_Elem ss
        cdef Topol tt
        cdef Poly* yy
        if len(kwrds) == 3:
            d = kwrds.pop("dim_type")
            s = kwrds.pop("spec_elem")
            t = kwrds.pop("topology")
            if isinstance(d, int):
                dd = d # needs to be a python int? Gotta figure out how to typed integer conversions. Ask about this. 
                ss = string_to_Spec_Elem(s)
                tt = string_to_Topol(t)
                self.thisptr = new Poly(dd, ss, tt)
                return
            raise ValueError("dim_type is required to be an instance of `:class:int`. spec_elem is either a string `universe` or `empty`, topology is either `nnc` or `closed`.")
        if "nnc_poly" in kwrds.keys():
            nnc_poly = kwrds.pop("nnc_poly")
            if isinstance(nnc_poly, NNC_Polyhedron):
                yy = (<NNC_Polyhedron> nnc_poly).thisptr
                self.thisptr = new Poly(yy[0])
                return
            raise TypeError(":class:`NNC_Polyhedron` needs to be provided to use the nnc_poly key word constructor.")
        if "cons" in kwrds.keys():
            cons = kwrds.pop("cons")
            d_cons = max([c.space_dimension() for c in cons])
            if "dim_type" in kwrds.keys():
                d_in = kwrds.pop("dim_type")
                d = max(d_cons, d_in)
            else:
                d = d_cons
            dd = d
            ss =  string_to_Spec_Elem("universe")
            
            tt = string_to_Topol("nnc")
            self.thisptr = new Poly(dd, ss, tt)
            for c in cons:
                cc = (<Constraint> c).thisptr[0] 
                self.thisptr.add_con(cc)
            return
        if "gens" in kwrds.keys():
            gens = kwrds.pop("gens")
            d_gens = max([g.space_dimension() for g in gens])
            if kwrds["dim_type"]:
                d_in = kwrds.pop("dim_type")
                d = max(d_gens, d_in)
            else:
                d = d_gens
            dd = d
            ss =  string_to_Spec_Elem("universe")
            tt = string_to_Topol("nnc")
            self.thisptr = new Poly(dd, ss, tt)
            for g in gens:
                gg = (<PPliteGenerator> g).thisptr[0] 
                self.thisptr.add_gen(gg)
            return
        raise KeyError("No valid keyword input for constructor of class given. ")

    def __cinit__(self):
        self.thisptr = NULL

    def __dealloc__(self):
        del self.thisptr

    def __hash__(self):
        return self.thisptr[0].hash()

    def __repr__(self):
        s = ""
        if self.is_empty(): 
            s = "false" # from pplite::Poly_Impl::print()
            return s
        self.minimize()
        comma = False
        for c in self.constraints():
            if comma:
                s += ", "
            s +=  str(c)
            comma = True
        return s

    def __eq__(self, other):
        if isinstance(other, NNC_Polyhedron):
            return self.equals(other)
        raise TypeError("Comparison with NNC polys only!")

    def is_necessarily_closed(self):
        return self.thisptr[0].is_necessarily_closed()

    def check_inv(self):
        return self.thisptr[0].check_inv()

    def is_empty(self):
        return self.thisptr[0].is_empty()

    def is_universe(self):
        return self.thisptr[0].is_universe()

    def is_minimized(self):
        return self.thisptr[0].is_minimized()

    def is_topologically_closed(self):
        return self.thisptr[0].is_topologically_closed()

    def is_bounded(self):
        return self.thisptr[0].is_bounded()

    def is_bounded_expression(self, from_below, expression):
        cdef Linear_Expr expr
        if isinstance(expression, Linear_Expression):
            expr = (<Linear_Expression> expression).thisptr[0]
        else:
            raise TypeError("expression needs to be of :class:`Linear_Expression`.")
        cdef cppbool f_b
        f_b = from_below
        return self.thisptr[0].is_bounded_expr(f_b, expr)

    def constrains(self,variable):
        cdef Var* vv
        if isinstance(variable, Variable):
            vv = (<Variable> variable).thisptr
            return self.thisptr[0].constrains(vv[0])
        raise TypeError("variable needs to be of :class:`Variable`.")

    def equals(self, other_poly):
        cdef Poly* yy
        if isinstance(other_poly, NNC_Polyhedron):
            yy = (<NNC_Polyhedron> other_poly).thisptr
            return self.thisptr[0].equals(yy[0])
        raise TypeError("other_poly needs to be of :class:`NNC_Polyhedron`")

    def contains(self, other_poly):
        cdef Poly* yy
        if isinstance(other_poly, NNC_Polyhedron):
            yy = (<NNC_Polyhedron> other_poly).thisptr
            return self.thisptr[0].contains(yy[0])
        raise TypeError("other_poly needs to be of :class:`NNC_Polyhedron`")

    def strictly_contains(self, other_poly):
        cdef Poly* yy
        if isinstance(other_poly, NNC_Polyhedron):
            yy = (<NNC_Polyhedron> other_poly).thisptr
            return self.thisptr[0].strictly_contains(yy[0])
        raise TypeError("other_poly needs to be of :class:`NNC_Polyhedron`")

    def is_disjoint_from(self, other_poly):
        cdef Poly* yy
        if isinstance(other_poly, NNC_Polyhedron):
            yy = (<NNC_Polyhedron> other_poly).thisptr
            return self.thisptr[0].is_disjoint_from(yy[0])
        raise TypeError("other_poly needs to be of :class:`NNC_Polyhedron`")

    def get_bounding_box(self):
        pass

    def boxed_contains(self, other_poly):
        cdef Poly* yy
        if isinstance(other_poly, NNC_Polyhedron):
            yy = (<NNC_Polyhedron> other_poly).thisptr
            return self.thisptr[0].boxed_contains(yy[0])
        raise TypeError("other_poly needs to be of :class:`NNC_Polyhedron`")

    # queries
    def topology(self):
        pass

    def space_dim(self):
        return self.thisptr.space_dim()

    def affine_dim(self):
        return self.thisptr[0].affine_dim()

    def _relation_with_c(self, constraint):
        cdef Con cc
        cdef Poly_Con_Rel p_c_r
        if isinstance(constraint, Constraint):
            cc = (<Constraint> constraint).thisptr[0]
            p_c_r = self.thisptr[0].relation_with(cc)
            result = Polyhedron_Constraint_Rel()
            result.thisptr[0] = p_c_r
            return result
        else:
            raise TypeError()

    def _relation_with_g(self, generator):
        cdef Gen gg
        cdef Poly_Gen_Rel p_g_r
        if isinstance(generator, PPliteGenerator):
            gg = (<PPliteGenerator> generator).thisptr[0]
            p_g_r = self.thisptr[0].relation_with(gg)
            result = Polyhedron_Generator_Rel()
            result.thisptr[0] = p_g_r
            return result
        else:
            raise TypeError("A :class:`PPliteGenerator` or a :class:`Constraint` should be passed into this method.")

    def relation_with(self, gen_or_constraint):
        try:
            return self._relation_with_c(gen_or_constraint)
        except TypeError:
            return self._relation_with_g(gen_or_constraint)  # failure here will raise the right type error for the general method. 



    def min(self, affine_expr, value, included_pointer, gen_object):
        cdef Affine_Expr ae
        cdef FLINT_Rational val
        cdef cppbool* included_ptr
        cdef Gen* g_ptr
        if isinstance(affine_expr, Affine_Expression):
            ae = (<Affine_Expression> affine_expr).thisptr[0]
        val = Python_float_to_FLINT_Rational(value)
        if isinstance(gen_object, PPliteGenerator):
            g_ptr = (<PPliteGenerator> gen_object).thisptr
        if isinstance(included_pointer, bool):
            included_ptr[0] = included_pointer # Doth this work?
        return self.thisptr[0].min(ae, val, included_ptr, g_ptr)

    def max(self, affine_expr, value, included_pointer, gen_object):
        cdef Affine_Expr ae
        cdef FLINT_Rational val
        cdef cppbool* included_ptr
        cdef Gen* g_ptr
        if isinstance(affine_expr, Affine_Expression):
            ae = (<Affine_Expression> affine_expr).thisptr[0]
        val = Python_float_to_FLINT_Rational(value)
        if isinstance(gen_object, PPliteGenerator):
            g_ptr = (<PPliteGenerator> gen_object).thisptr
        if isinstance(included_pointer, bool):
            included_ptr[0] = included_pointer # Doth this work?
        return self.thisptr[0].max(ae, val, included_ptr, g_ptr)

    def _get_bounds_v(self, variable):
        cdef Var* v
        cdef Itv itv 
        v = (<Variable> variable).thisptr
        itv = self.thisptr[0].get_bounds(v[0])
        i = Interval()
        i.interval = itv
        return i

    def _get_bounds_ae(self, affine_expr):
        cdef Affine_Expr ae
        cdef Itv itv 
        ae = (<Affine_Expression> affine_expr).thisptr[0]
        itv = self.thisptr[0].get_bounds(ae)
        i = Interval()
        i.interval = itv
        return i
        
    def _get_boundes_itv(self, itv_expr):
        pass

    def get_bounds(self, variable_or_affine_expr):
        if isinstance(variable_or_affine_expr, Variable):
            return self._get_bounds_v(variable_or_affine_expr)
        if isinstance(variable_or_affine_expr, Affine_Expression):
            return self._get_bounds_ae(variable_or_affine_expr)
        raise TypeError("A :class:`Variable` or a :class:`Affine_Expression` should be passed into this method.")

    def get_unconstrainted(self):
        pass

    def constraints(self):
        # Access constraints indirectly via copy_cons()
        # TODO: Properly implement via sys and Cons_Proxy in Poly_Impl
        cdef Cons constraint_vector 
        constraint_vector = self.thisptr[0].copy_cons()
        result = []
        cdef unsigned int index = constraint_vector.size() # hacky way to iterate over vectors
        for i in range(index):
            c = Constraint()
            c.thisptr = new Con(constraint_vector[i])
            result.append(c)
        return result

    def generators(self):
        """
        Returns a list of :class:`PPliteGenerator`.
        """
        # Access constraints indirectly via copy_cons()
        # TODO: Properly implement via sys and Cons_Proxy in Poly_Impl
        cdef Gens generator_vector 
        generator_vector = self.thisptr[0].copy_gens()
        result = []
        cdef unsigned int index = generator_vector.size()
        for i in range(index):
            g = PPliteGenerator()
            g.thisptr = new Gen(generator_vector[i])
            result.append(g)
        return result

    def normalized_constraints(self):
        # TODO implement once Cons_Proxy is implemented. 
        pass

    def num_min_constrains(self):
        return self.thisptr[0].num_min_cons()

    def num_min_generators(self):
        return self.thisptr[0].num_min_gens()

    def collapse(self, n):
        cdef dim_type nn
        if isinstance(n, int):
            nn = n
        self.thisptr[0].collapse(nn)

    def num_disjuncts(self):
        return self.thisptr[0].num_disjuncts()

    def disjunct_constraints(self, n):
        # TODO implement once Cons_Proxy is implemented. 
        # cdef dim_type nn
        # if isinstance(n, int):
        #     nn = n
        # cdef Cons_Proxy constraint_proxy
        pass

    def geom_covers(self, other_poly):
        """
        Input: :class:`NNC_Polyhedron`
        Output: bool
        """
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            return self.thisptr[0].geom_covers(y[0])

    def m_swap(self, other_poly):
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].m_swap(y[0])

    def set_empty(self):
        """
        Sets the :class:`NNC_Polyhedron` to empty. 
        """
        self.thisptr[0].set_empty()

    def set_topology(self, topology):
        cdef Topol tt
        tt = string_to_Topol(topology)
        self.thisptr[0].set_topology(tt)

    def add_constraint(self, constraint):  
        r"""
        TESTS::
        >>> from pplite import NNC_Polyhedron, Variable, Linear_Expression, Affine_Expression, Constraint, Point, Ray, Line, Closure_point
        >>> A = Variable(0)
        >>> P = NNC_Polyhedron(dim_type = 1, spec_elem = "universe", topology = "nnc")
        >>> P.add_constraint(A >= 0)
        >>> P
        x0>=0
        >>> P_2 = NNC_Polyhedron(nnc_poly = P)
        >>> P_2.add_constraint(A >= -10)
        >>> P_2.add_constraint(A >= -5)
        >>> P_2.add_constraint(A >= -2)
        >>> P_2.add_constraint(A >= -1)
        >>> P_2.add_constraint(A >= 0)
        >>> P_2.is_necessarily_closed()
        False
        >>> P_2.space_dim()
        1
        >>> P_2.equals(P)
        True
        >>> B = Variable(1)
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "empty", topology = "nnc") 
        >>> P_2 = NNC_Polyhedron(nnc_poly = P)
        >>> P.add_constraint(A == B)
        >>> P_2 == P
        True
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc") # addcons1 - test04
        >>> P.add_constraint(A >= 0)
        >>> P.add_constraint(A <= 2)
        >>> P.add_constraint(A >= -1)
        >>> P.add_constraint(B >= 1)
        >>> P_2 =  NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P_2.add_constraint(A >= 0)
        >>> P_2.add_constraint(A <= 2)
        >>> P_2.add_constraint(B >= 1)
        >>> P_2 == P
        True
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc") # addcons1 - test05
        >>> P.add_constraint(B >= 0)
        >>> P.add_constraint(B <= 2)
        >>> P.add_constraint(A + B >= 1)
        >>> P.add_constraint(A - B >= -1)
        >>> P_2 = NNC_Polyhedron(dim_type = 2, spec_elem = "empty", topology = "nnc")
        >>> P_2.add_generator(Point(B))
        >>> P_2.add_generator(Ray(A))
        >>> P_2.add_generator(Point(A+2*B))
        >>> P_2.add_generator(Point(A))
        >>> P == P_2
        True
        """
        if isinstance(constraint, Constraint):
            c = (<Constraint> constraint).thisptr[0]
            self.thisptr[0].add_con(c)
            return
        raise TypeError("Constraint not added. Check if input is a :class:`Constraint`.")


    def add_constraints(self, iter_of_cons):
        """
        TESTS::
        >>> from pplite import NNC_Polyhedron, Variable, Linear_Expression, Affine_Expression, Constraint
        >>> A = Variable(0)
        >>> B = Variable(1)
        >>> cons_list = [A >= 0, B == 5]
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc") #TODO: Make nicer python constructors
        >>> P.add_constraints(cons_list)
        >>> P_2 = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P_2.add_constraint(A >= 0)
        >>> P_2.add_constraint(B == 5)
        >>> P_2 == P
        True
        """
        for con in iter_of_cons:
            self.add_constraint(con)

    def add_generator(self, generator):
        """
        TESTS::
        >>> from pplite import NNC_Polyhedron, Variable, Linear_Expression, Affine_Expression, Constraint, Point, Closure_point, Ray, Line
        >>> A = Variable(0)
        >>> B = Variable(1)
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P.add_constraint(A >= 0)
        >>> P.add_constraint(B >= 0)
        >>> P.add_constraint(A + B > 0)
        >>> P.add_generator(Point())
        >>> P.minimize()
        >>> P_2 = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P_2.add_constraint(A >= 0)
        >>> P_2.add_constraint(B >= 0)
        >>> P_2.minimize()
        >>> P == P_2 # Test 01 finished
        True
        >>> P = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P.add_constraint(A >= 0)
        >>> P.add_constraint(B >= 0)
        >>> P.add_constraint(A + B > 0)
        >>> P.add_generator(Ray(-A))
        >>> P_2 = NNC_Polyhedron(dim_type = 2, spec_elem = "universe", topology = "nnc")
        >>> P_2.add_constraint(B >= 0)
        >>> P == P_2 # Test 03 finished
        True     
        """
        if isinstance(generator, PPliteGenerator):
            g = (<PPliteGenerator> generator).thisptr[0]
            self.thisptr[0].add_gen(g)

    def add_generators(self, iter_of_gens):
        for gen in iter_of_gens:
            self.add_generator(gen)

    def topological_closure_assign(self):
        self.thisptr[0].topological_closure_assign()

    def unconstain(self, variable):
        if isinstance(variable, Variable):
            v = (<Variable> variable).thisptr
            self.thisptr[0].unconstrain(v[0])

    def unconstain_many(self, iter_of_var_or_index_set):
        pass

    def intersection_assign(self, other_poly):
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].intersection_assign(y[0])

    def join_assign(self, other_poly):
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].join_assign(y[0])

    def poly_hull_assign(self, other_poly):
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].poly_hull_assign(y[0])

    def con_hull_assign(self, other_poly, boxed):
        cdef cppbool bboxed
        if boxed:
            bboxed = True
        else:
            bboxed = False
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].con_hull_assign(y[0], bboxed)

    def poly_difference_assign(self, other_poly):
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].poly_difference_assign(y[0])

    def affine_image(self, variable, linear_exp, inhomogenous_term, denominator):
        if isinstance(variable, Variable):
            var = (<Variable> variable).thisptr
        if isinstance(linear_exp, Linear_Expression):
            expr = (<Linear_Expression> linear_exp).thisptr[0]
        cdef FLINT_Integer inhomo
        cdef FLINT_Integer den
        inhomo = Python_int_to_FLINT_Integer(inhomogenous_term)
        den = Python_int_to_FLINT_Integer(denominator)
        self.thisptr[0].affine_image(var[0], expr, inhomo, den)

    def affine_preimage(self, variable, linear_exp, inhomogenous_term, denominator):
        if isinstance(variable, Variable):
            var = (<Variable> variable).thisptr
        if isinstance(linear_exp, Linear_Expression):
            expr = (<Linear_Expression> linear_exp).thisptr[0]
        cdef FLINT_Integer inhomo
        cdef FLINT_Integer den
        inhomo = Python_int_to_FLINT_Integer(inhomogenous_term)
        den = Python_int_to_FLINT_Integer(denominator)
        self.thisptr[0].affine_preimage(var[0], expr, inhomo, den)        
    
    # TODO: Implement these   
    def parallel_affine_image(self, args):
        pass

    def widing_assign(self, args):
        pass

    def time_elapse_assign(self, other_poly):
        if isinstance(other_poly, NNC_Polyhedron):
            y = (<NNC_Polyhedron> other_poly).thisptr
            self.thisptr[0].time_elapse_assign(y[0])

    def minimize(self):
        self.thisptr[0].minimize()

#####################################
### Poly_Con_Rel and Poly_Gen_Rel ###
#####################################

# TODO: Add full functionality of these classes.

cdef class Polyhedron_Constraint_Rel(object):
    def __cinit__(self):
        self.thisptr = NULL
    def __dealloc__(self):
        del self.thisptr

cdef class Polyhedron_Generator_Rel(object):
    def __cinit__(self):
        self.thisptr = NULL
    def __dealloc__(self):
        del self.thisptr


# TODO Migrate helper functions to a helper function module.

#########################
###  Helper Functions ###
#########################

cdef Topol string_to_Topol(t):
    cdef Topol tt
    if t == "closed":
        tt = Topol.CLOSED
        return tt
    if  t == "nnc":
        tt = Topol.NNC
        return tt
    raise ValueError("Topology type conversion failed.")

cdef Spec_Elem string_to_Spec_Elem(s):
    cdef Spec_Elem ss
    if s == "empty":
        ss = Spec_Elem.EMPTY
        return ss
    if  s == "universe":
        ss = Spec_Elem.UNIVERSE
        return ss
    raise ValueError("Spec_Elem type conversion failed.")


# distutils: extra_compile_args = -std=c++11

from .pplite_decl cimport *

cdef class Bounding_Box_t:
    cdef Box_t *thisptr

cdef class Bounding_Box_f:
    cdef Box_f *thisptr

# cpdef Create_Bounding_Box(cppbool keep_volume_info, dim_type sd=*, Spec_Elem se=*)


# distutils: extra_compile_args = -std=c++11

from __future__ import absolute_import

from .pplite_decl cimport *
####################################################

cdef FLINT_Integer_to_Python(FLINT_Integer& integer)

cdef FLINT_Integer Python_int_to_FLINT_Integer(integer)

cdef FLINT_Rational_to_Python(FLINT_Rational& rational)

cdef FLINT_Rational Python_float_to_FLINT_Rational(rational)


# distutils: language = c++
# distutils: libraries = gmp gmpxx pplite m flint

cimport cython

from gmpy2 cimport import_gmpy2, mpz, mpz_t, GMPy_MPZ_From_mpz, MPZ_Check, mpq, MPQ_Check, GMPy_MPQ_From_mpz
from libcpp.vector cimport vector as cppvector

# It is assumed that the pplite environment is set up to use FLINT_integers.
# TODO:  Write a proper conversion module to handle the Integer class in PPLite so this works regardless of setup.

import_gmpy2()

cdef FLINT_Integer_to_Python(FLINT_Integer& integer):
    """ Converts FLINT_Integer to python object.

    INPUT: :class:`FLINT_Interger` (c++)

    OUTPUT: Python object
    """
    cdef mpz_t new_int
    mpz_init(new_int)
    fmpz_get_mpz(new_int, integer.impl())
    y = GMPy_MPZ_From_mpz(new_int)
    mpz_clear(new_int)
    return y

cdef FLINT_Integer Python_int_to_FLINT_Integer(integer):
    """ Converts a python object to a c++ FLINT_Integer

    INPUT: 

    - integer: string or int

    OUTPUT:

    :class:`FLINT_Integer` (c++)
    """
    cdef fmpz_t x
    cdef fmpz y
    if isinstance(integer, (int, str)):
        fmpz_init(x)
        fmpz_set_si(x, integer)
    return FLINT_Integer(x)
    if MPZ_Check(integer): # is this okay?
        y = <fmpz> integer
        return FLINT_Integer(y)
    raise ValueError("Integer Conversion Failed")

cdef FLINT_Rational_to_Python(FLINT_Rational& rational):
    """Converts the Flint_Rational c++ class to a python object.

    INPUT:

    - rational: FLINT_Rational (c++)

    OUTPUT:

    - mpq

    """
    cdef mpz_t a
    cdef mpz_t b
    mpz_init(a)
    mpz_init(b)
    fmpq_get_mpz_frac(a , b, rational.impl())
    frac = GMPy_MPQ_From_mpz(a, b)
    mpz_clear(a)
    mpz_clear(b)
    return frac

cdef FLINT_Rational Python_float_to_FLINT_Rational(rational):
    """ Converts python float or fraction to a FLINT_Rational (c++).

    INPUT:

    - rational: object with method .as_integer_ratio()

    OUTPUT:

    FLINT_Rational (c++) 
    """
    cdef FLINT_Integer num
    cdef FLINT_Integer den
    try:
        numerator, denominator = rational.as_integer_ratio()
    except ValueError:
        raise ValueError("Rational Conversion Failed.")
    num = Python_int_to_FLINT_Integer(numerator)
    dem = Python_int_to_FLINT_Integer(denominator)
    return FLINT_Rational(num, dem)

def FLINT_Integer_Conversion_Check(possible_integer):
    """
    Checks a python object is convertible to a FLINT_Integer.

    Input: Object

    Output: Bool
    """
    if isinstance(possible_integer, (int, str)):
        return True
    if MPZ_Check(possible_integer):
        return True
    return False



# distutils: language = c++
# distutils: libraries = gmp gmpxx pplite m flint

cimport cython

from gmpy2 cimport import_gmpy2, mpz, mpz_t, GMPy_MPZ_From_mpz, MPZ_Check
# from libcpp.vector cimport vector as cppvector

from .integer_conversions cimport FLINT_Integer_to_Python, Python_int_to_FLINT_Integer, FLINT_Rational_to_Python, Python_float_to_FLINT_Rational
from .linear_algebra cimport Variable, Linear_Expression
from .constraint cimport Constraint
from .generators cimport PPliteGenerator
from .intervals cimport Interval

import_gmpy2()

cdef class Bounding_Box_t(object):
    """
    Wrapper for PPlite's BBox<true> class.

    Input: 

    Output: 
    
    TESTS:
    >>> from pplite.bounding_box import Bounding_Box_t
    >>> box_ex = Bounding_Box_t(1, "empty")
    >>> box_ex
    empty
    >>> box_ex_2 = Bounding_Box_t(box_ex); box_ex_2
    empty
    >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
    >>> x = Variable(0)
    >>> g = PPliteGenerator('ray', Linear_Expression(x), 1)
    >>> g_2 = PPliteGenerator('point', Linear_Expression(x), 0)
    >>> box_ex_2.add_generators([g,g_2]); box_ex_2 # Incorrect output, unsure why. might have something to do with interval class, conjecture either I"m not reading what should 
    # be defined here correctly. 
    x0 in [mpq(1,1), mpq(1,0)]
    # box isn't directly tested. it is tested indirectly via poly class. I should drop these tests. 

    """
    def __init__(self, *args): # special constructors as a class method. 
        if len(args) == 1:
            if isinstance(args[0], Bounding_Box_t):
                y = <Bounding_Box_t> args[0]
                self.thisptr = new Box_t(y.thisptr[0])
                return
            # if isinstance(other_box, Bounding_Box_f):
            #     y = <Bounding_Box_f> other_box
            #     self.thisptr = new Box_t(y.thisptr[0])
            #     return
        cdef dim_type sd
        cdef Spec_Elem se
        if len(args) == 2:
            sd = args[0]
            if args[1] == "empty":
                se = Spec_Elem.EMPTY
            elif args[1] == "universe":
                se = Spec_Elem.UNIVERSE
            else:
                raise ValueError("universe or empty box should be specified in second argument")
            self.thisptr = new Box_t(sd, se)
            return 

    def __cinit__(self):
        self.thisptr = NULL

    def __dealloc__(self):
        del self.thisptr

    def __repr__(self):
        s = ""
        if self.is_empty():
            s += "empty"
            return s
        else:
            for i in range(self.space_dimemsion()):
                s += str(Variable(i)) + " in " + str(self.get_bounds(Variable(i)))
                if i < self.space_dimemsion():
                    s += "\n"
            return s

    def __hash__(self):
        return self.thisptr.hash()

    def compute_volume_info(self):
        """
        Computes the volume of the Bounding Box.
        """
        cdef pair[dim_type, FLINT_Rational] Volume_Info
        Volume_Info = self.thisptr.compute_volume_info()
        cdef dim_type dim
        cdef FLINT_Rational vol
        dim = Volume_Info.first
        vol = Volume_Info.second
        py_vol = FLINT_Rational_to_Python(vol)
        return dim, py_vol

    def maybe_update_volume_info(self):
        self.thisptr.maybe_update_volume_info()

    def num_rays(self):
        """
        Returns the number of rays in the bounding box. 

        """
        return self.thisptr.num_rays()

    def pseudo_volume(self):
        """
        Returns the pseduo volume of the bounding box.
        """
        cdef FLINT_Rational p_vol
        p_vol = self.thisptr.pseudo_volume()
        return FLINT_Rational_to_Python(p_vol)

    def check_inv(self):
        return self.thisptr.check_inv()

    def space_dimemsion(self):
        return self.thisptr.space_dim()

    def affine_dimemsion(self):
        return self.thisptr.affine_dim()

    def is_empty(self):
        return self.thisptr.is_empty()

    def is_universe(self):
        return self.thisptr.is_universe()

    def is_bounded(self):
        return self.thisptr.is_bounded()

    def topology(self):
        pass

    def is_topologically_closed(self):
        return self.thisptr.is_topologically_closed()

    def constrains(self, variable):
        if isinstance(variable, Variable):
            vv = <Variable> variable
            return self.thisptr.constrains(vv.thisptr[0])

    def contains(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.contains(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.contains(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def equals(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.equals(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.equals(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def is_disjoint_from(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.is_disjoint_from(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.is_disjoint_from(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def less(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.less(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.less(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def number_of_minumum_constraints(self):
        return self.thisptr.num_min_cons()

    def number_of_minumum_generators(self):
        return self.thisptr.num_min_gens()

    def swap(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.swap(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.swap(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def set_empty(self):
        self.thisptr.set_empty()

    def set_origin(self):
        self.thisptr.set_origin()

    def affine_image(self, variable, expression, inhomogenous_term, denominator):
        cdef Var* var_ptr
        cdef Linear_Expr expr
        cdef FLINT_Integer inhomo
        cdef FLINT_Integer den
        # I want to use a try/except block but I havne't see code like that in sage and am unsure for this type ot type checking 
        # if the assumption that things are input correctly outweights teh patter of explicit checking. 
        if isinstance(variable, Variable):
            var_ptr = (<Variable> variable).thisptr
            if isinstance(expression, Linear_Expression):
                expr = (<Linear_Expression> expression).thisptr[0]
                inhom = Python_int_to_FLINT_Integer(inhomogenous_term)
                den = Python_int_to_FLINT_Integer(denominator)
                self.thisptr.affine_image(var_ptr[0], expr, inhomo, den)

    def concatenate_assign(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.concatenate_assign(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.concatenate_assign(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def glb_assign(self, other_box):
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.glb_assign(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def lub_assign(self, other_box):
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.lub_assign(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def unconstrain(self, variables):
        """
        INPUT: python iterable of :class:`Variable`
        """
        cdef Var* var
        for variable in variables:
            if isinstance(variable, Variable):
                var = (<Variable> variable).thisptr
                self.thisptr.unconstrain(var[0])

    def widening_assign(self, other_box):
        # if isinstance(other_box, Bounding_Box_f):
        #     w_o_vol_box = <Bounding_Box_f> other_box
        #     return self.thisptr.widening_assign(w_o_vol_box.thisptr[0])
        if isinstance(other_box, Bounding_Box_t):
            w_vol_box = <Bounding_Box_t> other_box
            return self.thisptr.widening_assign(w_vol_box.thisptr[0])

        raise ValueError("other_box needs to be a Bounding Box")

    def add_space_dims(self, dim, project):
        """
        Input: dim - int, project - bool
        """
        cdef dim_type d
        d = dim
        self.thisptr.add_space_dims(d, project)
        
    def remove_space_dim(self, variable):
        cdef Var* var
        if isinstance(variable, Variable):
            var = (<Variable> variable).thisptr
            self.thisptr.remove_space_dim(var[0])

    def remove_space_dims(self, variables):
        for variable in variables:
            self.remove_space_dim(variable)

    def remove_higher_space_dims(self, new_dimemsnion):
        cdef dim_type new_dim
        new_dim = new_dimemsnion
        self.thisptr.remove_higher_space_dims(new_dim)

    def expand_space_dim(self, variable, dimemsion_type):
        cdef Var* var
        cdef dim_type d
        if isinstance(variable, Variable):
            var = (<Variable> variable).thisptr
            d = dimemsion_type
            self.thisptr.expand_space_dim(var[0], d)

    def add_generator(self, generator):
        """
        """
        cdef Gen g
        if isinstance(generator, PPliteGenerator):
            g = (<PPliteGenerator> generator).thisptr[0]
            self.thisptr.add_gen(g)

    def add_generators(self, generators):
        for generator in generators:
            self.add_generator(generator)

    def refine_bounds(self, dimension, interval):
        cdef Itv y_itv
        if isinstance(interval, Interval):
            y_itv = (<Interval> interval).interval
            self.thisptr.refine_bounds(dimension, y_itv)

    def refine_as_intergral(self, *args):
        cdef dim_type dim
        cdef dim_type first
        cdef dim_type last
        if len(args) == 1:
            dim =  args[0]
            self.thisptr.refine_as_integral(dim)
        if len(args) == 2:
            first =  args[0]
            last = args[1]
            self.thisptr.refine_as_integral(first, last)            
        if len(args)>2:
            raise ValueError("expects 1 or 2 dimensions as input")

    def get_bounds(self, v):
        """
        """
        cdef Itv i
        if isinstance(v, Variable):
            vv = (<Variable> v).thisptr
            i = self.thisptr.get_bounds(vv[0])
            ii = Interval()
            ii.interval = i
            return ii

    def is_included_in(self, constraint):
        if isinstance(constraint, Constraint):
            cc = <Constraint> constraint
            return self.thisptr.is_included_in(cc.thisptr[0])

    def unconstrain_lb(self, dim):
        self.thisptr.unconstrain_lb(dim)

    def unconstrain_ub(self, dim):
        self.thisptr.unconstrain_ub(dim)

    def inf_lb(self, dim):
        return self.thisptr.inf_lb(dim)

    def inf_ub(self, dim):
        return self.thisptr.inf_ub(dim)

    def lb(self, dim):
        cdef FLINT_Rational q
        q = self.thisptr.lb(dim)
        return FLINT_Rational_to_Python(q)

    def ub(self, dim):
        cdef FLINT_Rational q
        q = self.thisptr.ub(dim)
        return FLINT_Rational_to_Python(q)

cdef class Bounding_Box_f(object):
    """
    Wrapper for PPlite's BBox<false> class.

    Input: 

    Output: 
    
    TESTS:
    >>> from pplite.bounding_box import Bounding_Box_f
    >>> box_ex = Bounding_Box_f(1, "empty")
    >>> box_ex
    empty
    >>> box_ex_2 = Bounding_Box_f(box_ex); box_ex_2
    empty
    >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
    >>> x = Variable(0)
    >>> g = PPliteGenerator('ray', Linear_Expression(x), 1)
    >>> g_2 = PPliteGenerator('point', Linear_Expression(x), 0)
    >>> box_ex_2.add_generators([g,g_2]); box_ex_2 # Incorrect output, unsure why. might have something to do with interval class, conjecture either I"m not reading what should 
    # be defined here correctly. 
    x0 in [mpq(1,1), mpq(1,0)]
    # box isn't directly tested. it is tested indirectly via poly class. I should drop these tests. 

    """
    def __init__(self, *args): # special constructors as a class method. 
        if len(args) == 1:
            # if isinstance(args[0], Bounding_Box_t):
            #     y = <Bounding_Box_t> args[0]
            #     self.thisptr = new Box_t(y.thisptr[0])
            #     return
            if isinstance(args[0], Bounding_Box_f):
                y = <Bounding_Box_f> args[0]
                self.thisptr = new Box_f(y.thisptr[0])
                return
        cdef dim_type sd
        cdef Spec_Elem se
        if len(args) == 2:
            sd = args[0]
            if args[1] == "empty":
                se = Spec_Elem.EMPTY
            elif args[1] == "universe":
                se = Spec_Elem.UNIVERSE
            else:
                raise ValueError("universe or empty box should be specified in second argument")
            self.thisptr = new Box_f(sd, se)
            return 

    def __cinit__(self):
        self.thisptr = NULL

    def __dealloc__(self):
        del self.thisptr

    def __repr__(self):
        s = ""
        if self.is_empty():
            s += "empty"
            return s
        else:
            for i in range(self.space_dimemsion()):
                s += str(Variable(i)) + " in " + str(self.get_bounds(Variable(i)))
                if i < self.space_dimemsion():
                    s += "\n"
            return s

    def __hash__(self):
        return self.thisptr.hash()

    def compute_volume_info(self):
        """
        Computes the volume of the Bounding Box.
        """
        cdef pair[dim_type, FLINT_Rational] Volume_Info
        Volume_Info = self.thisptr.compute_volume_info()
        cdef dim_type dim
        cdef FLINT_Rational vol
        dim = Volume_Info.first
        vol = Volume_Info.second
        py_vol = FLINT_Rational_to_Python(vol)
        return dim, py_vol

    def maybe_update_volume_info(self):
        self.thisptr.maybe_update_volume_info()

    def num_rays(self):
        """
        Returns the number of rays in the bounding box. 

        """
        return self.thisptr.num_rays()

    def pseudo_volume(self):
        """
        Returns the pseduo volume of the bounding box.
        """
        cdef FLINT_Rational p_vol
        p_vol = self.thisptr.pseudo_volume()
        return FLINT_Rational_to_Python(p_vol)

    def check_inv(self):
        return self.thisptr.check_inv()

    def space_dimemsion(self):
        return self.thisptr.space_dim()

    def affine_dimemsion(self):
        return self.thisptr.affine_dim()

    def is_empty(self):
        return self.thisptr.is_empty()

    def is_universe(self):
        return self.thisptr.is_universe()

    def is_bounded(self):
        return self.thisptr.is_bounded()

    def topology(self):
        pass

    def is_topologically_closed(self):
        return self.thisptr.is_topologically_closed()

    def constrains(self, variable):
        if isinstance(variable, Variable):
            vv = <Variable> variable
            return self.thisptr.constrains(vv.thisptr[0])

    def contains(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.contains(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.contains(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def equals(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.equals(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.equals(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def is_disjoint_from(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.is_disjoint_from(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.is_disjoint_from(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def less(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.less(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.less(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def number_of_minumum_constraints(self):
        return self.thisptr.num_min_cons()

    def number_of_minumum_generators(self):
        return self.thisptr.num_min_gens()

    def swap(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.swap(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.swap(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def set_empty(self):
        self.thisptr.set_empty()

    def set_origin(self):
        self.thisptr.set_origin()

    def affine_image(self, variable, expression, inhomogenous_term, denominator):
        cdef Var* var_ptr
        cdef Linear_Expr expr
        cdef FLINT_Integer inhomo
        cdef FLINT_Integer den
        # I want to use a try/except block but I havne't see code like that in sage and am unsure for this type ot type checking 
        # if the assumption that things are input correctly outweights teh patter of explicit checking. 
        if isinstance(variable, Variable):
            var_ptr = (<Variable> variable).thisptr
            if isinstance(expression, Linear_Expression):
                expr = (<Linear_Expression> expression).thisptr[0]
                inhom = Python_int_to_FLINT_Integer(inhomogenous_term)
                den = Python_int_to_FLINT_Integer(denominator)
                self.thisptr.affine_image(var_ptr[0], expr, inhomo, den)

    def concatenate_assign(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.concatenate_assign(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.concatenate_assign(w_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def glb_assign(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.glb_assign(w_o_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def lub_assign(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.lub_assign(w_o_vol_box.thisptr[0])
        raise ValueError("other_box needs to be a Bounding Box")

    def unconstrain(self, variables):
        """
        INPUT: python iterable of :class:`Variable`
        """
        cdef Var* var
        for variable in variables:
            if isinstance(variable, Variable):
                var = (<Variable> variable).thisptr
                self.thisptr.unconstrain(var[0])

    def widening_assign(self, other_box):
        if isinstance(other_box, Bounding_Box_f):
            w_o_vol_box = <Bounding_Box_f> other_box
            return self.thisptr.widening_assign(w_o_vol_box.thisptr[0])
        # if isinstance(other_box, Bounding_Box_t):
        #     w_vol_box = <Bounding_Box_t> other_box
        #     return self.thisptr.widening_assign(w_vol_box.thisptr[0])

        raise ValueError("other_box needs to be a Bounding Box")

    def add_space_dims(self, dim, project):
        """
        Input: dim - int, project - bool
        """
        cdef dim_type d
        d = dim
        self.thisptr.add_space_dims(d, project)
        
    def remove_space_dim(self, variable):
        cdef Var* var
        if isinstance(variable, Variable):
            var = (<Variable> variable).thisptr
            self.thisptr.remove_space_dim(var[0])

    def remove_space_dims(self, variables):
        for variable in variables:
            self.remove_space_dim(variable)

    def remove_higher_space_dims(self, new_dimemsnion):
        cdef dim_type new_dim
        new_dim = new_dimemsnion
        self.thisptr.remove_higher_space_dims(new_dim)

    def expand_space_dim(self, variable, dimemsion_type):
        cdef Var* var
        cdef dim_type d
        if isinstance(variable, Variable):
            var = (<Variable> variable).thisptr
            d = dimemsion_type
            self.thisptr.expand_space_dim(var[0], d)

    def add_generator(self, generator):
        """
        """
        cdef Gen g
        if isinstance(generator, PPliteGenerator):
            g = (<PPliteGenerator> generator).thisptr[0]
            self.thisptr.add_gen(g)

    def add_generators(self, generators):
        for generator in generators:
            self.add_generator(generator)

    def refine_bounds(self, dimension, interval):
        cdef Itv y_itv
        if isinstance(interval, Interval):
            y_itv = (<Interval> interval).interval
            self.thisptr.refine_bounds(dimension, y_itv)

    def reinfe_as_integral(self, *args):
        cdef dim_type dim
        cdef dim_type first
        cdef dim_type last
        if len(args) == 1:
            dim =  args[0]
            self.thisptr.refine_as_integral(dim)
        if len(args) == 2:
            first =  args[0]
            last = args[1]
            self.thisptr.refine_as_integral(first, last)            
        if len(args)>2:
            raise ValueError("expects 1 or 2 dimensions as input")

    def get_bounds(self, v):
        """
        """
        cdef Itv i
        if isinstance(v, Variable):
            vv = (<Variable> v).thisptr
            i = self.thisptr.get_bounds(vv[0])
            ii = Interval()
            ii.interval = i
            return ii

    def is_included_in(self, constraint):
        if isinstance(constraint, Constraint):
            cc = <Constraint> constraint
            return self.thisptr.is_included_in(cc.thisptr[0])

    def unconstrain_lb(self, dim):
        self.thisptr.unconstrain_lb(dim)

    def unconstrain_ub(self, dim):
        self.thisptr.unconstrain_ub(dim)

    def inf_lb(self, dim):
        return self.thisptr.inf_lb(dim)

    def inf_ub(self, dim):
        return self.thisptr.inf_ub(dim)

    def lb(self, dim):
        cdef FLINT_Rational q
        q = self.thisptr.lb(dim)
        return FLINT_Rational_to_Python(q)

    def ub(self, dim):
        cdef FLINT_Rational q
        q = self.thisptr.ub(dim)
        return FLINT_Rational_to_Python(q)


# cpdef Create_Bounding_Box(keep_volume_info, arg_0 = None, arg_1 = None):
#     """
#     Wrapper for PPlite's BBox class. 

#     Input: Book keep_volue_info, 

#     Output: 

#     TESTS:

#     """
#     if arg_0 is None and arg_1 is None:
#         raise ValueError("At ")
#     if keep_volume_info == True:
#         return Bounding_Box_t(arg_0, arg_1)
#     if keep_volume_info == False:
#         return Bounding_Box_f(arg_0, arg_1)
# # #  


# distutils: extra_compile_args = -std=c++11

from __future__ import absolute_import

from .pplite_decl cimport *

cdef class NNC_Polyhedron(object):
    cdef Poly* thisptr

cdef class Polyhedron_Constraint_Rel(object):
    cdef Poly_Con_Rel* thisptr

cdef class Polyhedron_Generator_Rel(object):
    cdef Poly_Gen_Rel* thisptr



# cdef class Constraint_Proxy(object):
#     cdef Cons_Proxy c_proxy


# distutils: language = c++
# distutils: libraries = gmp gmpxx pplite m flint

cimport cython

from gmpy2 cimport import_gmpy2, mpz, mpz_t, GMPy_MPZ_From_mpz, MPZ_Check
from libcpp.vector cimport vector as cppvector
from .integer_conversions cimport FLINT_Integer_to_Python, Python_int_to_FLINT_Integer, FLINT_Rational_to_Python, Python_float_to_FLINT_Rational
from .constraint cimport Constraint
from .linear_algebra cimport Variable

cdef class Interval(object):
    """
    cython wrapper for pplite ``Itv`` struct. 
    This represent a topologically closed 1 dimensional interval.

    EXAMPLES:
    >>> from pplite.intervals import Interval
    >>> I = Interval()
    >>> I["lb"] = 3/2
    >>> I["ub"] = 5/2
    >>> I
    [mpq(3,2), mpq(5,2)]
    >>> J = Interval()
    >>> J
    (-inf, +inf)

    Intervals are mutable:

    >>> I["lb"] = 4/2
    >>> I
    [mpq(2,1), mpq(5,2)]
    >>> I["lb"] = 100
    >>> I # this is a bug
    [mpq(100,1), mpq(5,2)] 
    >>> I.is_empty() # something is wrong here. investigate
    False
    """
    def __hash__(self):
        cdef size_t h
        h = self.interval.hash()
        return h

    def __repr__(self):
        """
        TESTS:
        """
        s = ""
        if self.is_universe():
            s += "(-inf, +inf)"
            return s
        elif self.inf_lb() and self.has_ub():
            s += "(-inf, "+self["ub"].__repr__()+"]"
            return s
        elif self.inf_ub() and self.has_lb():
            s += "["+self["lb"].__repr__()+", +inf)"
            return s
        elif self.has_ub() and self.has_lb():
            s += "["+self["lb"].__repr__()+", "+self["ub"].__repr__()+"]"
            return s
        elif self.is_empty():
            s += "{ }"
            return s

    def __setitem__(self, member, value):
        """
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["lb"]
        mpq(3,2)
        >>> I["ub"] = 12
        >>> I["ub"]
        mpq(12,1)
        """
        if member == "lb":
            self.set_lower_bound(value)
        if member == "ub":
            self.set_upper_bound(value)
        if member == "singlton":
            self.set_singleton(value)


    def __getitem__(self, member):
        """
        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["lb"]
        mpq(3,2)
        >>> I["ub"] = 12
        >>> I["ub"]
        mpq(12,1)
        """
        if member == "lb":
            return self.get_lower_bound()
        if member == "ub":
            return self.get_upper_bound()

    def set_lower_bound(self, value):
        """
        Sets the lower bound of the interval. 

        INPUT: Value - Rational or value convertible to python rational. 

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2 #  indirect test
        >>> I["lb"]
        mpq(3,2)
        """
        cdef FLINT_Rational lower_bound
        lower_bound = Python_float_to_FLINT_Rational(value)
        self.interval.set_lb(lower_bound)

    def get_lower_bound(self):
        """
        Gets the lower bound of the interval. 

        OUTPUT: Rational

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2 #  indirect test
        >>> I["lb"]
        mpq(3,2)
        """
        cdef FLINT_Rational lower_bound
        lower_bound = self.interval.lb
        return FLINT_Rational_to_Python(lower_bound)

    def set_upper_bound(self, value):
        """
        Sets the upper bound of the interval. 

        INPUT: Value - Rational or value convertible to python rational. 

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["ub"] = 3/2 #  indirect test
        >>> I["ub"]
        mpq(3,2)
        """
        cdef FLINT_Rational upper_bound
        upper_bound = Python_float_to_FLINT_Rational(value)
        self.interval.set_ub(upper_bound)

    def get_upper_bound(self):
        """
        Gets the upper bound of the interval. 

        OUTPUT: Rational

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["ub"] = 3/2 #  indirect test
        >>> I["ub"]
        mpq(3,2)
        """
        cdef FLINT_Rational upper_bound
        upper_bound  = self.interval.ub
        return FLINT_Rational_to_Python(upper_bound)

    def check_inv(self):
        """
        Calls the check inv method in the interval struct.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 12
        >>> I.check_inv()
        True
        """
        return self.interval.check_inv()

    def is_empty(self):
        """
        Checks if the interval is empty.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 12
        >>> I.is_empty()
        False
        >>> J = Interval()
        >>> J.set_empty()
        >>> J.is_empty()
        True
        """
        return self.interval.is_empty()

    def is_universe(self):
        """
        Checks if the interval is universe (i.e. the real line).

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 12
        >>> I.is_universe()
        False
        >>> J = Interval()
        >>> J.set_universe()
        >>> J.is_universe()
        True
        """
        return self.interval.is_universe()

    def has_lb(self):
        """
        Checks if the interval has a lower bound.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I.has_lb()
        True
        >>> J = Interval()
        >>> J["ub"] = 12
        >>> J.has_lb()
        False
        """
        return self.interval.has_lb()

    def has_ub(self):
        """
        Checks if the interval has an upper bound.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I.has_ub()
        False
        >>> J = Interval()
        >>> J["ub"] = 12
        >>> J.has_ub()
        True
        """
        return self.interval.has_ub()

    def inf_lb(self):
        """
        Checks if the interval is unbounded from below.
        Output: Bool

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I.inf_lb()
        True
        """
        return self.interval.inf_lb()

    def inf_ub(self):
        """
        Checks if the interval is unbounded from above.
        Output: Bool

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I.inf_ub()
        True
        """
        return self.interval.inf_ub()

    def is_bounded(self):
        """
        Checks if the interval has an upper bound.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I.has_lb()
        True
        >>> J = Interval()
        >>> J["ub"] = 12
        >>> J.has_lb()
        False
        """
        return self.interval.is_bounded()

    def is_singleton(self):
        """
        Checks if the interval has an upper bound.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 3/2
        >>> I.is_singleton()
        True
        >>> J = Interval()
        >>> J["ub"] = 3/2
        >>> J["ub"] = 12
        >>> J.is_singleton()
        False
        """
        return self.interval.is_singleton()

    def is_zero(self):
        """
        Checks if the interval is zero.

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 12
        >>> I.is_zero()
        False
        >>> J = Interval()
        >>> J.set_zero()
        >>> J.is_zero()
        True
        """
        return self.interval.is_zero()

    def is_disjoint_from(self, other_interval):
        """
        Determines if interval I is disjoint form interval J. 

        INPUT: other_interval a :class:`Interval`

        OUTPUT: Bool
        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 12
        >>> J = Interval()
        >>> J.set_zero()
        >>> I.is_disjoint_from(J)
        True
        >>> J["ub"] = 10
        >>> I.is_disjoint_from(J)
        False
        """
        cdef Itv y
        if isinstance(other_interval, Interval):
            y = (<Interval> other_interval).interval
            return self.interval.is_disjoint_from(y)
        raise ValueError("other_interval is required to be an Interval")

    def intersects(self, other_interval):
        """
        Determines if interval I is intersects interval J. 

        INPUT: other_interval a :class:`Interval`

        OUTPUT: Bool

        TESTS: 
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 12
        >>> J = Interval()
        >>> J.set_zero()
        >>> I.intersects(J)
        False
        >>> J["ub"] = 10
        >>> I.intersects(J)
        True
        """
        cdef Itv y
        if isinstance(other_interval, Interval):
            y = (<Interval> other_interval).interval
            return self.interval.intersects(y)
        raise ValueError("other_interval is required to be an Interval")

    def length(self):
        """
        Computes the length of an interval.

        OUTPUT: :class:`mpq`

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I["lb"] = 3/2
        >>> I["ub"] = 5/2
        >>> I.length()
        mpq(1,1)
        """
        cdef FLINT_Rational l 
        l = self.interval.length()
        return FLINT_Rational_to_Python(l)

    def num_min_cons(self):
        """

        """
        cdef dim_type d
        d = self.interval.num_min_cons()
        return d

    def num_rays(self):
        """
        Returns the number of rays.
        """
        cdef dim_type d
        d = self.interval.num_rays()
        return d

    def set_empty(self):
        """
        Sets the interval to the empty set.

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval(); I
        (-inf, +inf)
        >>> I.set_empty()
        >>> I
        { }
        """
        self.interval.set_empty()

    def set_universe(self):
        """
        Sets the interval to the empty set.

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval();
        >>> I["ub"] = 12; I["lb"] = 2; I
        [mpq(2,1), mpq(12,1)]
        >>> I.set_universe()
        >>> I
        (-inf, +inf)
        """
        self.interval.set_universe()

    def set_zero(self):
        """
        Sets the interval to the empty set.

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I
        (-inf, +inf)
        >>> I.set_zero()
        >>> I
        [mpq(0,1), mpq(0,1)]
        """
        self.interval.set_zero()

    def set_singleton(self, value):
        """
        Sets the interval to a singleton of the specified value.

        INPUT: rational

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval()
        >>> I
        (-inf, +inf)
        >>> I.set_singleton(3/4)
        >>> I
        [mpq(3,4), mpq(3,4)]
        """
        cdef FLINT_Rational v
        v = Python_float_to_FLINT_Rational(value)
        self.interval.set_singleton(v)

    def unset_lower_bound(self):
        """
        Unset the interval's lower bound value.

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval();
        >>> I["ub"] = 12; I["lb"] = 2; I
        [mpq(2,1), mpq(12,1)]
        >>> I.unset_lower_bound()
        >>> I
        (-inf, mpq(12,1)]
        """
        self.interval.unset_lb()

    def unset_upper_bound(self):
        """
        Unset the interval's upperbound bound value.

        TESTS:
        >>> from pplite.intervals import Interval
        >>> I = Interval();
        >>> I["ub"] = 12; I["lb"] = 2; I
        [mpq(2,1), mpq(12,1)]
        >>> I.unset_upper_bound()
        >>> I
        [mpq(2,1), +inf)
        """
        self.interval.unset_ub()

# Double check what these functions should do. 

    def glb_assign(self, other_interval):
        cdef Itv y
        if isinstance(other_interval, Interval):
            y = (<Interval> other_interval).interval
            return self.interval.glb_assign(y)
        raise ValueError("other_interval must be an Interval.")

    def lub_assign(self, other_interval):
        cdef Itv y
        if isinstance(other_interval, Interval):
            y = (<Interval> other_interval).interval
            self.interval.lub_assign(y)
        raise ValueError("other_interval must be an Interval.")            

    def widen_assign(self, other_interval):
        cdef Itv y
        if isinstance(other_interval, Interval):
            y = (<Interval> other_interval).interval
            self.interval.widen_assign(y)
        raise ValueError("other_interval must be an Interval.")

    def refine_as_integral(self):
        return self.interval.refine_as_integral()

    def complement_assign(self):
        self.interval.complement_assign()

    def add_assign(self, other_interval):
        cdef Itv y
        if isinstance(other_interval, Interval):
            y = (<Interval> other_interval).interval
            self.interval.add_assign(y)
        raise ValueError("other_interval must be an Interval.")

    def mul_assign(self, value):
        cdef FLINT_Rational r
        r = Python_float_to_FLINT_Rational(value)
        self.interval.mul_assign(r)

###########################
###########################
### Functions in itv.hh ###
###########################
###########################

def interval_from_con_inhomo(constraint):
    cdef Con c
    cdef Itv result
    if isinstance(constraint, Constraint):
        c = (<Constraint> constraint).thisptr[0]
        result = itv_from_con_inhomo(c)
        new_interval = Interval()
        new_interval.interval = result
        return new_interval
    raise ValueError("constraint must be a Constraint.")

def interval_from_itv_con(constraint):
    cdef Con c
    cdef Itv result
    if isinstance(constraint, Constraint):
        c = (<Constraint> constraint).thisptr[0]
        result = itv_from_itv_con(c)
        new_interval = Interval()
        new_interval.interval = result
        return new_interval
    raise ValueError("constraint must be a Constraint.")

def split_interval(this_interval, constraint, intergral):
    """
    INPUT: this_interval :class:`Interval`, constraint :class:`Constraint`, integral :bool:
    OUTPUT: :class:`Interval`
    """
    cdef Itv itv
    cdef Con c
    cdef cppbool b
    cdef Itv result
    if isinstance(this_interval, Interval):
        itv = (<Interval> this_interval).interval
    if isinstance(constraint, Constraint):
        c = (<Constraint> constraint).thisptr[0]
    if intergral is True:
        b = True
    else:
        b = False
    result = split_itv(itv, c, b)
    new_interval = Interval()
    new_interval.interval = result
    return new_interval

def get_lower_bound_constraint(variable, this_interval):
    cdef Var* var
    cdef Itv itv
    cdef Con result
    if isinstance(variable, Variable):
        var = (<Variable> variable).thisptr
    else:
        var = (<Variable> Variable(variable)).thisptr
    if isinstance(this_interval, Interval):
        itv = (<Interval> this_interval).interval
    else:
        raise ValueError("this_interval is required to be an Interval")
    result = get_lb_con(var[0], itv)
    new_constraint = Constraint()
    new_constraint.thisptr[0] = result
    return new_constraint

def get_upper_bound_constraint(variable, this_interval):
    cdef Var* var
    cdef Itv itv
    cdef Con result
    if isinstance(variable, Variable):
        var = (<Variable> variable).thisptr
    else:
        var = (<Variable> Variable(variable)).thisptr
    if isinstance(this_interval, Interval):
        itv = (<Interval> this_interval).interval
    else:
        raise ValueError("this_interval is required to be an Interval")
    result = get_ub_con(var[0], itv)
    new_constraint = Constraint()
    new_constraint.thisptr[0] = result
    return new_constraint

def get_equality_constraint(variable, this_interval):
    cdef Var* var
    cdef Itv itv
    cdef Con result
    if isinstance(variable, Variable):
        var = (<Variable> variable).thisptr
    else:
        var = (<Variable> Variable(variable)).thisptr
    if isinstance(this_interval, Interval):
        itv = (<Interval> this_interval).interval
    else:
        raise ValueError("this_interval is required to be an Interval")
    result = get_eq_con(var[0], itv)
    new_constraint = Constraint()
    new_constraint.thisptr[0] = result
    return new_constraint


# distutils: language = c++
# distutils: libraries = gmp gmpxx pplite m flint

from gmpy2 cimport GMPy_MPZ_From_mpz, import_gmpy2, mpz, mpz_t, GMPy_MPZ_From_mpz, MPZ_Check
from cython.operator cimport dereference as deref
from .linear_algebra import Variable, Linear_Expression, Affine_Expression
from .integer_conversions cimport FLINT_Integer_to_Python, Python_int_to_FLINT_Integer

cdef class PPliteGenerator(object):
    """
    Wrapper for PPlite's Gen class.

    Input: :class:`PPliteGenerator`, or Generator Type, :class:`Linear_Experssion` and `int`.

    Output: :class:`PPlite_Generator`
    """
    def __init__(self, *args):
        """
        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> PPliteGenerator()
        p(0)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('line', e, 7); g
        l(2*x0+3*x1)
        >>> g_2 = PPliteGenerator(g); g_2
        l(2*x0+3*x1)
        >>> e_2 = 2*x-3*y
        >>> g_3 =  PPliteGenerator('point', e_2, 9); g_3
        p((2*x0-3*x1)/9)
        >>> g_4 = PPliteGenerator('ray', Linear_Expression(x), 0); g_4
        r(x0)
        """
        if len(args) == 0:
            self.thisptr = new Gen()
            return
        cdef Gen gg
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, PPliteGenerator): 
                gg = (<PPliteGenerator> arg).thisptr[0]
                self.thisptr = new Gen(gg)
                return
            raise ValueError("creating an instance of PPliteGenerator from one input requires the input to be a PPliteGenerator.")
        cdef GenType tt
        cdef FLINT_Integer dd
        cdef Linear_Expr ee 
        if len(args) == 3:
            t = args[0]
            e = args[1]
            d = args[2]
            tt = string_to_GenType(t)
            if isinstance(e, Linear_Expression):
                ee = (<Linear_Expression> e).thisptr[0]
                dd = Python_int_to_FLINT_Integer(d)
                self.thisptr = new Gen(tt, ee, dd)
                return
            raise ValueError("Three inputs requires a Generator Type, linear expression, and integer")
        raise ValueError("Four or more arguments cannot be used to define a generator.")

    def __dealloc__(self):
        """
        The cython deconstructor.
        """
        del self.thisptr

    def __repr__(self):
        """
        Representation of a generator in the space of generators. 

        For example p(x0+x1) corresponds to the point (1,1) as an ordered pair. 

        OUTPUT::

        string
        """
        need_div = False
        s = ""
        if self.gen_type() == "line":
            s += "l("
        elif self.gen_type() == "ray":
            s += "r("
        elif self.gen_type() == "point":
            s += "p("
            if self.divisor() != 1:
                need_div = True
        elif self.gen_type() == "closure_point":
            s += "c("
            if self.divisor() != 1:
                need_div = True
        if need_div:
            s += "("
        s += str(self.linear_form())
        if need_div:
            s += ")/"
            s += str(self.divisor())
        s += ")"
        return s

    def __hash__(self):
        raise TypeError("PPliteGenerator unhashable")

    def __eq__(self, other_gen):
        cdef Gen y
        if isinstance(other_gen, PPliteGenerator):
            y =  (<PPliteGenerator> other_gen).thisptr[0]
            return self.thisptr[0] == y

    def gen_type(self):
        """
        Returns the type of generator. 

        Output: "line", "ray", "point", or "closure_point"

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('line', e, 7)
        >>> g.gen_type()
        'line'
        """
        cdef GenType t
        t = self.thisptr.type()
        if t == GenType.LINE:
            return "line"
        if t == GenType.RAY:
            return "ray"
        if t == GenType.POINT:
            return "point"
        if t == GenType.CLOSURE_POINT:
            return "closure_point"

    def set_type(self, t):
        """
        Sets the generator type of a generatorg

        Output: "line", "ray", "point", or "closure_point"

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('line', e, 7)
        >>> g.gen_type()
        'line'
        >>> g.set_type("ray")
        >>> g.gen_type()
        'ray'
        """
        cdef GenType tt
        tt = string_to_GenType(t)
        self.thisptr.set_type(tt)

    def is_line(self):
        """
        Returns true if the generator is a line. 

        Output: bool

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('line', e, 7)
        >>> g.is_line()
        True
        >>> g = PPliteGenerator('ray', e, 7)
        >>> g.is_line()
        False
        """
        return self.thisptr.is_line()

    def is_ray(self):
        """
        Returns true if the generator is a ray. 

        Output: bool

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('line', e, 7)
        >>> g.is_ray()
        False
        >>> g = PPliteGenerator('ray', e, 7)
        >>> g.is_ray()
        True
        """
        return self.thisptr.is_ray()

    def is_point(self):
        """
        Returns true if the generator is a point. 

        Output: bool

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.is_point()
        True
        >>> g = PPliteGenerator('closure_point', e, 7)
        >>> g.is_point()
        False
        """
        return self.thisptr.is_point()

    def is_closure_point(self):
        """
        Returns true if the generator is a closure point. 

        Output: bool

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.is_closure_point()
        False
        >>> g = PPliteGenerator('closure_point', e, 7)
        >>> g.is_closure_point()
        True
        """
        return self.thisptr.is_closure_point()

    def is_line_or_ray(self):
        """
        Returns true if the generator is a line or ray. 

        Output: bool

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.is_line_or_ray()
        False
        >>> g = PPliteGenerator('line', e, 7)
        >>> g.is_line_or_ray()
        True
        >>> g = PPliteGenerator('ray', e, 7)
        >>> g.is_line_or_ray()
        True
        """
        return self.thisptr.is_line_or_ray()

    def is_point_or_closure_point(self):
        """
        Returns true if the generator is a point or closure point. 

        Output: bool

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.is_point_or_closure_point()
        True
        >>> g = PPliteGenerator('closure_point', e, 7)
        >>> g.is_point_or_closure_point()
        True
        >>> g = PPliteGenerator('ray', e, 7)
        >>> g.is_point_or_closure_point()
        False
        """
        return self.thisptr.is_point_or_closure_point()

    def linear_form(self):
        """
        Returns the Linear Form associated with the generator.

        Output: :class:`Linear_Expression`

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.linear_form()
        2*x0+3*x1
        """
        cdef Linear_Expr e
        e = self.thisptr.linear_expr()
        expr = Linear_Expression()
        expr.thisptr[0] = e
        return expr

    def coefficient(self, v):
        """
        Returns the coefficient of variable for a given generator. 

        Input: `int` or :class:`Variable`

        Output: `int`

        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.coefficient(x)
        mpz(2)
        """
        cdef Var* vv
        if isinstance(v, Variable):
            vv = (<Variable> v).thisptr
        else:
            vv = (<Variable> Variable(v)).thisptr
        cdef FLINT_Integer n
        n = self.thisptr.coeff(vv[0])
        return FLINT_Integer_to_Python(n)

    def divisor(self):
        """
        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.divisor()
        mpz(7)
        """
        cdef FLINT_Integer n
        n = self.thisptr.divisor()
        return FLINT_Integer_to_Python(n)

    def space_dimision(self):
        """
        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> g = PPliteGenerator('point', e, 7)
        >>> g.space_dimision()
        2
        """
        return self.thisptr.space_dim()

    def set_space_dimension(self, dim_type dim):
        self.thisptr.set_space_dim(dim)

    def is_equal_to(self, y):
        """
        Tests if self is equal to y.
        TESTS:
        >>> from pplite import Variable, Linear_Expression, Affine_Expression, PPliteGenerator
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = 2*x + 3*y
        >>> e_2 = x + y
        >>> g_1 = PPliteGenerator('point', e, 7)
        >>> g_2 = PPliteGenerator('ray', e, 7)
        >>> g_1.is_equal_to(g_2)
        False
        >>> g_3 = PPliteGenerator('point', e, 7)
        >>> g_1.is_equal_to(g_3)
        True
        >>> g_4 = PPliteGenerator('line', e_2, 4)
        >>> g_4.is_equal_to(g_3)
        False
        """
        cdef Gen gg
        if isinstance(y, PPliteGenerator):
            gg = (<PPliteGenerator> y).thisptr[0]
            return self.thisptr.is_equal_to(gg)
        raise ValueError("This is for comparing two generators.")

    def check_inv(self):
        return self.thisptr.check_inv()

    def m_swap(self, y):
        cdef Gen gg
        if isinstance(y, PPliteGenerator):
            gg = (<PPliteGenerator> y).thisptr[0]
            self.thisptr.m_swap(gg)
            return
        raise ValueError("This works only when using an input of a generator")

    def permute_space_dims(self, cycle, d):
        return NotImplemented

    def shift_space_dims(self, dim_type start, dim_type n):
        self.thisptr.shift_space_dims(start, n)

    def is_line_or_equality(self):
        return self.thisptr.is_line_or_equality()

    def sign_normalize(self):
        self.thisptr.sign_normalize()

    def strong_normalize(self):
        self.thisptr.strong_normalize()

###########################
### Auxiliary Functions ###
###########################

# Provides equivalent functionality to line, ray, point, closure_point functions in gen.hh

def Point(linear_expression=Linear_Expression(), integer=1):
    """
    Constructs a point.
    """
    return PPliteGenerator("point", Linear_Expression(linear_expression), integer)

def Closure_point(linear_expression=Linear_Expression(), integer=1):
    """
    Constructs a closure point.
    """
    return PPliteGenerator("closure_point", Linear_Expression(linear_expression), integer)

def Line(linear_expression):
    """
    Constructs a line.
    """
    return PPliteGenerator("line", Linear_Expression(linear_expression), 0)

def Ray(linear_expression):
    """
    Constructs a ray.
    """
    return PPliteGenerator("ray", Linear_Expression(linear_expression), 0)
    

########################
### Helper Functions ###
########################

# TODO migrate helper functions to a single module. 

cdef GenType string_to_GenType(t):
    """
    Converts a string to an enum pplite::GenType.

    Input: String - 'line', 'ray', 'point', 'closure point'
    
    Output: GenType of the same name.
    """
    cdef GenType tt
    if t == "line":
        tt = GenType.LINE
        return tt
    if t == "ray":
        tt = GenType.RAY
        return tt
    if t == "point":
        tt = GenType.POINT
        return tt
    if t == "closure_point":
        tt = GenType.CLOSURE_POINT
        return tt
    raise ValueError("Input must be one of the following string: `line`, `ray`, `point`, `closure point`.")


# distutils: extra_compile_args = -std=c++11

from .linear_algebra cimport *

cdef _wrap_Constraint(Con constraint)

cdef _make_Constraint_from_richcmp(lhs_, rhs_, op)

cdef class Constraint(object):
    cdef Con *thisptr

cdef enum ConType:
    EQUALITY
    NONSTRICT_INEQUALITY
    STRICT_INEQUALITY


# distutils: language = c++
# distutils: libraries = gmp gmpxx pplite m flint

cimport cython

from gmpy2 cimport import_gmpy2, mpz, mpz_t, GMPy_MPZ_From_mpz, MPZ_Check
from libcpp.vector cimport vector as cppvector
from .constraint cimport _make_Constraint_from_richcmp
# from .integer_conversions cimport FLINT_Integer_to_Python, Python_int_to_FLINT_Integer

# TODO: Investigate why everything breaks when importing integer conversion as opposed to local definitions. 

import_gmpy2()

cdef FLINT_Integer_to_Python(FLINT_Integer& integer):
    r""" Converts PPLite::FLINT_Integer to python integer."""
    cdef mpz_t new_int
    mpz_init(new_int)
    fmpz_get_mpz(new_int, integer.impl())
    y = GMPy_MPZ_From_mpz(new_int)
    mpz_clear(new_int)
    return y

cdef FLINT_Integer Python_int_to_FLINT_Integer(integer):
    r"""Converts python sting or int to a PPLite::FLINT_Integer."""
    cdef fmpz_t x
    cdef fmpz y
    if isinstance(integer, (int, str)):
        fmpz_init(x)
        fmpz_set_si(x, integer)
        return FLINT_Integer(x)
    if MPZ_Check(integer): # is this okay?
        y = <fmpz> integer
        return FLINT_Integer(y)
    raise ValueError("Integer Conversion Failed")

################
### Variable ###
################

@cython.freelist(256)
cdef class Variable(object):
    r"""
    Wrapper for PPLites's ``Var`` class.

    INPUT:

    - ``i`` -- integer. The index of the axis.

    OUTPUT:

    A :class:`Variable`

    Examples:

    >>> from pplite import Variable
    >>> x = Variable(123)
    >>> x.id()
    123
    >>> x
    x123

    Note that the "meaning" of an object of the class Variable is completely
    specified by the integer index provided to its constructor: be careful not
    to be mislead by C++ language variable names. For instance, in the following
    example the linear expressions ``e1`` and ``e2`` are equivalent, since the
    two variables ``x`` and ``z`` denote the same Cartesian axis:

    >>> x = Variable(0)
    >>> y = Variable(1)
    >>> z = Variable(0)
    >>> e1 = x + y; e1
    x0+x1
    >>> e2 = y + z; e2
    x0+x1
    >>> e1 - e2
    0
    """
    def __cinit__(self, dim_type i):
        """
        The Cython constructor.

        See :class:`Variable` for documentation.

        Tests:

        >>> from pplite import Variable
        >>> Variable(123)   # indirect doctest
        x123
        """
        self.thisptr = new Var(i)

        
    def __dealloc__(self):
        """
        The Cython destructor.
        """
        del self.thisptr
        
    def __hash__(self):
        r"""
        Tests:

        >>> import pplite
        >>> hash(pplite.Variable(12))
        Traceback (most recent call last):
        ...
        TypeError: Variable unhashable
        """
        raise TypeError('Variable unhashable')

    def id(self):
        """
        Return the index of the Cartesian axis associated to the variable.

        Examples:

        >>> from pplite import Variable
        >>> x = Variable(123)
        >>> x.id()
        123
        """
        return self.thisptr.id()
        
    def space_dimension(self):
        r"""
        Return the dimension of the vector space enclosing ``self``.

        OUTPUT:

        Integer. The returned value is ``self.id()+1``.

        Examples:

        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> x.space_dimension()
        1
        """
        return self.thisptr.space_dim()


    def swap(self, Variable w):
        """
        Swaps representation of variables.
        """
        swap(self.thisptr[0], w.thisptr[0])
        return self
    
    def __repr__(self):
        """
        Return a string representation.

        OUTPUT:

        String.

        Examples:

        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> x.__repr__()
        'x0'
        """
        return 'x{0}'.format(self.id())
    
    def __add__(self, other):
        r"""
        Return the sum ``self`` + ``other``.

        INPUT:

        - ``self``, ``other`` -- anything convertible to
          ``Affine_Expression``: An integer, a :class:`Variable`,
          :class:`Linear_Expression`, or :class`Affine_Expression`.

        OUTPUT:

        A :class:`Linear_Expression`.

        Examples:

        >>> from pplite import Variable, Linear_Expression, Affine_Expression
        >>> x = Variable(0);
        >>> y = Variable(1)
        >>> x + y
        x0+x1
        >>> isinstance(x+y, Linear_Expression)
        True
        >>> isinstance(x+y, Affine_Expression)
        False
        """
        if not isinstance(self, Variable):
            raise NotImplemented 
        if not isinstance(other, Variable): # promote variable to linear expression. 
            return Linear_Expression(self) +  other
        # only use this method if both classes are an instance of Variable
        other_var = <Variable> other
        result = Linear_Expression()
        result.thisptr = new Linear_Expr(self.thisptr[0] + other_var.thisptr[0])
        return result

    def __radd__(self, other):
        """
        Return the sum ``other`` + ``self``.

        INPUT:

        - ``self``, ``other`` -- anything convertible to
          ``Affine_Expression``: An integer, a :class:`Variable`,
          :class:`Linear_Expression`, or :class`Affine_Expression`.

        OUTPUT:

        A :class:`Linear_Expression`. 
        """
        if not isinstance(self, Variable):
            raise NotImplemented
        if not isinstance(other, Variable):
            return other + Linear_Expression(self)
        other_var = <Variable> other
        result = Linear_Expression()
        result.thisptr = new Linear_Expr(self.thisptr[0] + other_var.thisptr[0])
        return result

    def __sub__(self, other):
        r"""
        Return the difference ``self`` - ``other``.

        INPUT:

        - ``self``, ``other`` -- anything convertible to
          ``Linear_Expression``: An integer, a :class:`Variable`, or a
          :class:`Linear_Expression`.

        OUTPUT:

        A :class:`Linear_Expression` representing ``self`` - ``other``.

        Examples:

        >>> from pplite import Variable
        >>> x = Variable(0); y = Variable(1)
        >>> x - y
        x0-x1
        """
        if not isinstance(self, Variable):
            return NotImplemented
        if not isinstance(other, Variable):
            return Linear_Expression(self) - other
        other_var = <Variable> other
        result = Linear_Expression()
        result.thisptr = new Linear_Expr(self.thisptr[0] - other_var.thisptr[0])
        return result


    def __rsub__(self, other):
        """
        Return the difference ``other`` - ``self``.

        INPUT:

        - ``self``, ``other`` -- anything convertible to
          ``Linear_Expression``: An integer, a :class:`Variable`, or a
          :class:`Linear_Expression`.

        OUTPUT:

        A :class:`Linear_Expression` representing ``self`` - ``other``.
        """
        if not isinstance(self, Variable): 
            return NotImplemented
        if not isinstance(other, Variable):
            return other - Linear_Expression(self)
        other_var = <Variable> other
        result = Linear_Expression()
        result.thisptr = new Linear_Expr(other_var.thisptr[0] - self.thisptr[0])
        return result

    def __mul__(self, other):
        r"""
        Return the product ``self`` * ``other``.

        INPUT:

        - ``self``, ``other`` -- One must be an integer, the other a
          :class:`Variable`.

        OUTPUT:

        A :class:`Linear_Expression` representing ``self`` * ``other``.

        Examples:

        >>> from pplite import Variable, Linear_Expression
        >>> x = Variable(0); y = Variable(1)
        >>> x * 15
        15*x0
        >>> e = 15 * y; e
        15*x1
        >>> isinstance(e, Linear_Expression)
        True
        >>> x * 1.5
        Traceback (most recent call last):
        ...
        ValueError: Integer Conversion Failed
        """
        if isinstance(self, Variable): # pplite doesn't explictly multiply Var and ints, convert to linear expr
            return Linear_Expression(self) * other
        return NotImplemented

    def __rmul__(self, other):
        return Linear_Expression(self) * other

    def __pos__(self):
        r"""
        Return ``self`` as :class:`Linear_Expression`

        OUTPUT:

        The :class:`Linear_Expression` ``+self``

        Examples:

        >>> from pplite import Variable
        >>> x = Variable(0); x
        x0
        >>> +x
        x0
        """
        return Linear_Expression(self)

    def __neg__(self):
        r"""
        Return -``self`` as :class:`Linear_Expression`

        OUTPUT:

        The :class:`Linear_Expression` ``-self``

        Examples:

        >>> from pplite import Variable
        >>> x = Variable(0); x
        x0
        >>> -x
        -x0
        """
        return Linear_Expression(self)*(-1)

    def __richcmp__(self, other, op):
        """
        Construct :class:`Constraint` from equalities or inequalities.

        INPUT:

        - ``self``, ``other`` -- anything convertible to a
          :class:`Linear_Expression`

        - ``op`` -- the operation.

        Examples:
        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> x <  y
        -x0+x1>0
        >>> x <= 0
        -x0>=0
        >>> x == y-y
        x0==0
        >>> x >= -2
        x0+2>=0
        >>> x >  0
        x0>0
        >>> 0 == 1  # watch out!
        False
        >>> 0*x == 1
        -1==0
        """
        return _make_Constraint_from_richcmp(self, other, op)

#########################
### Linear_Expression ###
#########################

cdef class Linear_Expression(object):
    r"""
    Wrapper for PPLite's ``Linear_Expr`` class.

    This class might more aptly be described as linear form rather than a linear expression. 
    For translation purposes, the class is named linear Expression to align with 
    the original pplite code and ppl.

    INPUT:

    The constructor accepts zero, one, or two arguments.

    If there are two arguments ``Linear_Expression(a,b)``, they are
    interpreted as

    - ``a`` -- a :class:`Linear_Expression`.

    - ``b`` -- an positive integer. The space dimension of a linear form.

    A single argument ``Linear_Expression(arg)`` is interpreted as

    - ``arg`` -- something that determines a linear
      expression. Possibilities are:

      * a :class:`Variable`: The linear expression given by that
        variable.

      * a :class:`Linear_Expression`: The copy constructor.

      * an integer: Constructs the 0 linear expression for space dimension of the integer.

    No argument is the default constructor and returns the zero linear
    expression.

    OUTPUT:

    A :class:`Linear_Expression`

    Examples:

    >>> from pplite import Variable, Linear_Expression

    >>> e = -3*Variable(1) + Variable(7); e
    -3*x1+x7
    >>> isinstance(e, Linear_Expression)
    True
    >>> e.space_dimension()
    8
    >>> e_2 = Linear_Expression(e, 20); e_2
    -3*x1+x7
    >>> e_2.space_dimension()
    20
    >>> Linear_Expression()
    0
    >>> e = Linear_Expression(5); e
    0
    >>> e.space_dimension()
    5
    >>> e = Linear_Expression(Variable(2)); e
    x2
    >>> x = Variable(123)
    >>> y = Variable(321)
    >>> expr = x+y
    >>> expr
    x123+x321
    >>> expr.coefficient(x)
    mpz(1)
    >>> expr.coefficient(Variable(124))
    mpz(0)

    String, rationals and floating point types are accepted as long as they
    represent exact integers:
    """
    # TODO: Finish Tests/claim above.
    def __init__(self, *args):
        """
        The Cython constructor.

        See :class:`Linear_Expression` for documentation.

        """
        cdef dim_type dim 
        if len(args) == 2:
            a = args[0]
            b = args[1]
            dim = b
            if isinstance(a, Linear_Expression):
                e = <Linear_Expression> a
                self.thisptr = new Linear_Expr(e.thisptr[0], b)
            return 
        elif len(args) == 1:
            arg = args[0]
            if isinstance(arg, Variable):
                v = <Variable> arg
                self.thisptr = new Linear_Expr(v.thisptr[0])
                return
            if isinstance(arg, Linear_Expression):
                e = <Linear_Expression> arg
                self.thisptr = new Linear_Expr(e.thisptr[0])
                return
            if isinstance(arg, int):
                dim = arg
                self.thisptr = new Linear_Expr(dim)
                return
            raise ValueError("Initalizing with one argument requires either a linear expression, variable, or integer to be passed in.")
        elif len(args) == 0:
            self.thisptr = new Linear_Expr()
            return
        else:
            raise ValueError("Cannot initialize with more than 2 arguments.")

    def __dealloc__(self):
        """
        The Cython destructor.
        """
        del self.thisptr

    def __hash__(self):
        r"""
        Tests:

        >>> import pplite
        >>> hash(pplite.Linear_Expression(10))
        Traceback (most recent call last):
        ...
        TypeError: Linear_Expression unhashable
        """
        raise TypeError('Linear_Expression unhashable')

    def space_dimension(self):
        """
        Return the dimension of the vector space necessary for the
        linear expression.

        OUTPUT:

        Integer.

        Examples::

        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> (x+y).space_dimension()
        2
        >>> (x+y).space_dimension()
        2
        >>> (y).space_dimension()
        2
        >>> (x).space_dimension()
        1
        >>> (y-y).space_dimension()
        2
        """
        return self.thisptr.space_dim()
    
    def set_space_dimension(self, dim_type dim):
        """
        Sets the ambient dimension which the linear expression lives in.

        INPUT: 

        - ``dim`` an integer
        """
        self.thisptr.set_space_dim(dim)

    def coefficient(self, v):
        """
        Return the coefficient of the variable ``v``.

        INPUT:

        - ``v`` -- a :class:`Variable` or an integer.

        OUTPUT:

        An (Python) Integer. 

        Examples::

        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> e = 3*x
        >>> e.coefficient(x)
        mpz(3)
        >>> e.coefficient(Variable(1))
        mpz(0)
        """
        cdef Var* vv
        cdef FLINT_Integer result
        if isinstance(v, Variable):
            vv = (<Variable> v).thisptr
            result = self.thisptr[0].get(vv[0])
            return FLINT_Integer_to_Python(result)
        cdef dim_type dim
        if isinstance(v, int):
            dim = v
            result = self.thisptr[0].get(dim)
            return FLINT_Integer_to_Python(result)
    
    def set_coefficient(self, i, n):
        """
        Set the ``i``-th coefficient to ``n``.

        INPUT:

        - ``i`` - variable or variable index

        - ``n`` - integer

        Examples::

        >>> from pplite import Variable
        >>> L = Variable(0) + (3 * Variable(1)); L
        x0+3*x1
        >>> L.set_coefficient(1, -5)
        >>> L
        x0-5*x1
        >>> L.set_coefficient(3, 7); L
        x0-5*x1+7*x3
        """
        cdef FLINT_Integer nn
        if isinstance(i, Variable):
            ii = <Variable> i
        else:
            var_i = Variable(i)
            ii = <Variable> var_i
        nn = Python_int_to_FLINT_Integer(n)
        (<Linear_Expression> self).thisptr[0].set(ii.thisptr[0], nn) 

    def __repr__(self):
        r"""
        Return a string representation of the linear expression.

        OUTPUT:

        A string.

        Examples::

        >>> from pplite import Linear_Expression, Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> x
        x0
        >>> x-x
        0
        >>> 2*x
        2*x0
        """
        s = ''
        first = True
        for i in range(self.space_dimension()):
            x = Variable(i)
            coeff = self.coefficient(x)
            if coeff == 0:
                continue
            if first and coeff == 1:
                s += '%r' % x
                first = False
            elif first and coeff == -1:
                s += '-%r' % x
                first = False
            elif first and coeff != 1:
                s += '%d*%r' % (coeff, x)
                first = False
            elif coeff == 1:
                s += '+%r' % x
            elif coeff == -1:
                s += '-%r' % x
            else:
                s += '%+d*%r' % (coeff, x)
        if first:
            s = '0'
        return s

    def swap_space_dimensions(self, v1, v2):
        r"""
        Swaps the coefficients of ``v1`` and ``v2``.

        INPUT:

        - ``v1``, ``v2`` - variables or indices of variables

        Examples:

        >>> from pplite import Variable
        >>> L = Variable(1) - 3 * Variable(3); L
        x1-3*x3
        >>> L.swap_space_dimensions(Variable(1), Variable(3))
        >>> L
        -3*x1+x3
        >>> L = Variable(1) - 3 * Variable(3)
        >>> L.swap_space_dimensions(1, 3)
        >>> L
        -3*x1+x3
        """
        cdef dim_type var_1, var_2
        if isinstance(v1, Variable):
            var_1 = v1.id()
        else:
            vv1 = new Var(v1)
            var_1 = vv1.id()
        if isinstance(v2, Variable):
            var_2 = v2.id()
        else:
            vv2 = new Var(v2)
            var_2 = vv2.id()
        self.thisptr.swap_space_dims(var_1, var_2)

    def shift_space_dimensions(self, v, dim_type n):
        r"""
        Shift by ``n`` the coefficients of variables starting from the
        coefficient of ``v``.

        This increases the space dimension by ``n``.

        INPUT:

        - ``v`` a :class:`Variable`

        - ``n`` an integer

        Examples::

        >>> from pplite import Variable
        >>> L = Variable(0) + 13 * Variable(2) + 5 * Variable(7)
        >>> L
        x0+13*x2+5*x7
        >>> L.shift_space_dimensions(Variable(2), 2)
        >>> L
        x0+13*x4+5*x9
        >>> L.shift_space_dimensions(Variable(7), 3)
        >>> L
        x0+13*x4+5*x12
        """
        cdef Variable vv
        if type(v) is Variable:
            vv = <Variable> v
        else:
            vv = Variable(v)
        self.thisptr.shift_space_dims(vv.thisptr[0], n)

    # def remove_space_dimensions(self, Variables_Set V):
    #     r"""
    #     Removes the dimension specified by the set of variables ``V``.

    #     See :class:`Variables_Set` to construct set of variables.

    #     Examples:

    #     >>> from pplite import Variable
    #     >>> L = sum(i * Variable(i) for i in range(10))
    #     >>> L
    #     x1+2*x2+3*x3+4*x4+5*x5+6*x6+7*x7+8*x8+9*x9
    #     >>> L.remove_space_dimensions(Variables_Set(3,5))
    #     >>> L
    #     x1+2*x2+6*x3+7*x4+8*x5+9*x6
    #     """
    #     self.thisptr.remove_space_dimensions(V.thisptr[0])

    def all_homogeneous_terms_are_zero(self):
        """
        Test if ``self`` is a constant linear expression.

        OUTPUT:

        Boolean.

        Examples:

        >>> from pplite import Variable, Linear_Expression
        >>> x = Variable(1)
        >>> (x-x).all_homogeneous_terms_are_zero()
        True
        """
        return self.thisptr.is_zero()

    def is_equal_to(self, Linear_Expression other):
        """
        Test equality with another linear expression.

        OUTPUT: boolean

        Examples::

        >>> from pplite import Variable
        >>> L1 = Variable(0) + 2 * Variable(3)
        >>> L2 = Variable(0) + 2 * Variable(3)
        >>> L3 = Variable(0) - Variable(2)
        >>> L1.is_equal_to(L2)
        True
        >>> L1.is_equal_to(L3)
        False
        """
        return self.thisptr.is_equal_to(other.thisptr[0])
        
    def __add__(self, other):
        r"""
        Add ``self`` and ``other``.

        INPUT:

        - ``self``, ``other`` -- anything that can be used to
          construct a :class:`Linear_Expression` or :class:`Affine_Expression`. One of them, not
          necessarily ``self``, is guaranteed to be a
          :class:``Linear_Expression``, otherwise Python would not
          have called this method.

        OUTPUT:

        The sum as a :class:`Linear_Expression` or :class:`Affine_Expression` depending on input.

        Examples::

        >>> from pplite import Affine_Expression, Linear_Expression, Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> x + y + y + y
        x0+3*x1
        >>> e = x + y + x
        >>> isinstance(e, Linear_Expression)
        True
        >>> e2 = 7 + e; e2
        2*x0+x1+7
        >>> isinstance(e2, Linear_Expression)
        False
        >>> isinstance(e2, Affine_Expression)
        True
        >>> e + 7
        2*x0+x1+7
        >>> e + Variable(0)
        3*x0+x1
        >>> Variable(1) + e
        2*x0+2*x1
        >>> e3 = e + e2; e3
        4*x0+2*x1+7
        >>> isinstance(e3, Affine_Expression)
        True
        """
        if not isinstance(self, Linear_Expression):
            return NotImplemented
        # to mimic pplite, we use cases for type conversions.
        # case 1: linear_expr + var -> linear_expr
        if isinstance(other, Variable):
            other_var = <Variable> other
            result = Linear_Expression()
            result.thisptr = new Linear_Expr(other_var.thisptr[0] + self.thisptr[0])  
            return result
        # case 2: linear_expr + linear_expr -> linear_expr
        cdef Linear_Expr* lhs_expr
        cdef Linear_Expr* rhs_expr
        cdef Linear_Expr result_expr
        if isinstance(other, Linear_Expression):
            lhs_expr = (<Linear_Expression> self).thisptr
            rhs_expr = (<Linear_Expression> other).thisptr
            result_expr = lhs_expr[0] + rhs_expr[0]
            result = Linear_Expression()
            result.thisptr[0] = result_expr
            return result
        # case 3: linear_expr + affine_expr -> affine_expr
        if isinstance(other, Affine_Expression):
            return NotImplemented 
        # case 4, linear_expr + integer -> affine expression
        result_aff = Affine_Expression(self, other)
        return result_aff

    def __radd__(self, other):
        return self + other        

    def __sub__(self, other):
        r"""
        Subtract ``self`` and ``other``.

        INPUT:

        - ``self``, ``other`` -- anything that can be used to
          construct a :class:`Linear_Expression` or :class:`Affine_Expression`. One of them, not
          necessarily ``self``, is guaranteed to be a
          :class:``Linear_Expression``, otherwise Python would not
          have called this method.

        OUTPUT:

        The sum as a :class:`Linear_Expression` or :class:`Affine_Expression` depending on input.

        Examples:

        >>> from pplite import Affine_Expression, Linear_Expression, Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> x - y + y + y
        x0+x1
        >>> e = y - x + y - x; e
        -2*x0+2*x1
        >>> isinstance(e, Linear_Expression)
        True
        >>> e - e
        0
        >>> e1 = y - x
        >>> e1 - e
        x0-x1
        >>> e2 = e - 7; e2
        -2*x0+2*x1-7
        >>> e3 = 7 - e; e3
        2*x0-2*x1+7
        >>> isinstance(e3, Linear_Expression)
        False
        >>> isinstance(e3, Affine_Expression)
        True
        """
        # to mimic pplite, we use cases for type conversions.
        # case 1: linear_expr - var -> linear_expr
        if not isinstance(self, Linear_Expression):
            return NotImplemented
        if isinstance(other, Variable):
            other_var = <Variable> other
            result = Linear_Expression()
            self.thisptr[0] -= other_var.thisptr[0]
            result.thisptr = new Linear_Expr(self.thisptr[0])  
            return result
        # case 2: linear_expr + linear_expr -> linear_expr
        cdef Linear_Expr* lhs_expr
        cdef Linear_Expr* rhs_expr
        cdef Linear_Expr result_expr
        if isinstance(other, Linear_Expression):
            lhs_expr = (<Linear_Expression> self).thisptr
            rhs_expr = (<Linear_Expression> other).thisptr
            result_expr = lhs_expr[0] - rhs_expr[0]
            result = Linear_Expression()
            result.thisptr[0] = result_expr
            return result
        # case 3: linear_expr + affine_expr -> affine_expr
        if isinstance(other, Affine_Expression):
            return NotImplemented 
        # case 4, linear_expr + integer -> affine expression
        result_aff = Affine_Expression(self, -other)
        return result_aff

    def __rsub__(self, other):
        if not isinstance(self, Linear_Expression):
            return NotImplemented
        if isinstance(other, Variable):
            other_var = <Variable> other
            result = Linear_Expression()
            result.thisptr = new Linear_Expr(other_var.thisptr[0] - self.thisptr[0])  
            return result
        cdef Linear_Expr* lhs_expr
        cdef Linear_Expr* rhs_expr
        cdef Linear_Expr result_expr
        if isinstance(other, Linear_Expression):
            lhs_expr = (<Linear_Expression> self).thisptr
            rhs_expr = (<Linear_Expression> other).thisptr
            result_expr = rhs_expr[0] - lhs_expr[0]
            result = Linear_Expression()
            result.thisptr[0] = result_expr
            return result
        if isinstance(other, Affine_Expression):
            return NotImplemented
        result_aff = Affine_Expression(-1*self, other)
        return result_aff

    def __mul__(self, other):
        r"""
        Multiply ``self`` with ``other``.

        INPUT:

        - ``self``, ``other`` -- anything that can be used to
          construct a :class:`Linear_Expression`. One of them, not
          necessarily ``self``, is guaranteed to be a
          :class:``Linear_Expression``, otherwise Python would not
          have called this method.

        OUTPUT:

        The product as a :class:`Linear_Expression`

        Examples::

        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> 8 * (x)
        8*x0
        >>> y * 8
        8*x1
        """
        #         >>> 2**128 * x
        # 340282366920938463463374607431768211456*x0
        #         >>> from gmpy2 import mpz
        # >>> mpz(3) * x * mpz(5)
        # 15*x0
        cdef Linear_Expr* e
        if isinstance(self, Linear_Expression):
            e = (<Linear_Expression> self).thisptr
            c = other
        else:
            # NOTE: this code path will only be executed when compiled with cython < 3.0.0
            e = (<Linear_Expression> other).thisptr
            c = self
        cdef FLINT_Integer cc = Python_int_to_FLINT_Integer(c)
        cdef Linear_Expression result = Linear_Expression()
        result.thisptr[0] = e[0] * cc
        return result

    def __rmul__(self, other):
        cdef Linear_Expr* e
        if isinstance(self, Linear_Expression):
            e = (<Linear_Expression> self).thisptr
            c = other
        else:
            # NOTE: this code path will only be executed when compiled with cython < 3.0.0
            e = (<Linear_Expression> other).thisptr
            c = self
        cdef FLINT_Integer cc = Python_int_to_FLINT_Integer(c)
        cdef Linear_Expression result = Linear_Expression()
        result.thisptr[0] = e[0] * cc
        return result

    def __neg__(self):
        neg_assign(self.thisptr[0])
        return self

    def __pos__(self):
        return self

    def __richcmp__(self, other, op):
        """
        Construct :class:`Constraint`s

        Examples::

        >>> from pplite import Variable
        >>> A = Variable(0)
        >>> B = Variable(1)
        >>> A > B
        x0-x1>0
        """
        return _make_Constraint_from_richcmp(self, other, op)


#########################
### Affine_Expression ###
#########################

cdef class Affine_Expression(object):
    r"""
    Wrapper for PPLite's ``Affine_Expr`` class.

    Note, this behaves analgolously to ppl's ``Linear_Expression`` class. 

    Examples:

    The constructor accepts zero, one, or two arguments.

    If there are two arguments ``Affine_Expression(a,b)``, they are
    interpreted as

    - ``a`` -- either a dictionary whose indices are space dimension and
      values are coefficients, an iterable coefficients (e.g. a list or
      tuple), or an :class:`Linear_Expression`.

    - ``b`` -- an integer. The inhomogeneous term.

    A single argument ``Affine_Expression(arg)`` is interpreted as

    - ``arg`` -- something that determines a affine
      expression. Possibilities are:

      * a :class:`Affine_Expression`: The copy constructor.

      * an integer: Constructs the constant affine expression.

    No argument is the default constructor and returns the zero affine
    expression.

    OUTPUT:

    A :class:`Affine_Expression`

    Examples::

    >>> from pplite import Variable, Linear_Expression, Affine_Expression
    >>> e = Variable(2) - 3*Variable(4)
    >>> isinstance(e, Linear_Expression)
    True
    >>> a = Affine_Expression(e, 7); a
    x2-3*x4+7
    >>> Affine_Expression()
    0
    >>> -a
    -x2+3*x4-7
    """
    def __init__(self, *args):
        """
        The Cython constructor.

        See :class:`Affine_Expression` for documentation.
        """
        cdef FLINT_Integer k
        if len(args) == 2:
            expr_arg = args[0]
            int_arg = args[1]   
            k = Python_int_to_FLINT_Integer(int_arg)
            if isinstance(expr_arg, Linear_Expression):
                e = <Linear_Expression> expr_arg
                self.thisptr = new Affine_Expr(e.thisptr[0], k)
                return
            raise ValueError("Initializing with two arguments requires a :class:`Linear_Expression` and an integer to be passed in.")
        if len(args) == 1:
            arg = args[0]   
            if isinstance(arg, int):
                k = Python_int_to_FLINT_Integer(arg)
                self.thisptr = new Affine_Expr(k)
                return
            if isinstance(arg, Affine_Expression):
                a = <Affine_Expression> arg 
                self.thisptr = new Affine_Expr(a.thisptr[0])
                return
            raise ValueError("Initializing with one argument requires either an :class:`Affine_Expression` xor an integer to be passed in.")
        elif len(args) == 0:
            self.thisptr = new Affine_Expr()
            return
        else:
            raise ValueError("Cannot initialize with more than 2 arguments.")

    def __dealloc__(self):
        """
        The Cython destructor.
        """
        del self.thisptr

    def __hash__(self):
        r"""
        Tests::

        >>> import pplite
        >>> hash(pplite.Affine_Expression(10))
        Traceback (most recent call last):
        TypeError: Affine_Expression unhashable
        """
        raise TypeError('Affine_Expression unhashable')

    def space_dimension(self):
        """
        Return the dimension of the vector space necessary for the
        linear expression.

        OUTPUT:

        Integer.

        Examples::

        >>> from pplite import Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> (x+y+1).space_dimension()
        2
        """
        return self.thisptr.space_dim()

    def linear_form(self):
        """
        Returns the linear form of the affine expression.

        OUTPUT:

        :class:`Linear_Expression`

        Examples::

        >>> from pplite import Variable, Affine_Expression
        >>> e = Variable(2) - 3*Variable(4)
        >>> a = Affine_Expression(e, 7); a
        x2-3*x4+7
        >>> a.linear_form()
        x2-3*x4
        """
        cdef Linear_Expr e
        e = (<Affine_Expression> self).thisptr[0].expr
        ee = Linear_Expression()
        ee.thisptr[0] = e
        return ee

    def inhomogeneous_term(self):
        """
        Returns the inhogogenous term of an affine expression.

        OUTPUT:

        Integer.

        Examples::

        >>> from pplite import Variable, Affine_Expression
        >>> e = Variable(2) - 3*Variable(4)
        >>> a = Affine_Expression(e, 7); a
        x2-3*x4+7
        >>> a.inhomogeneous_term()
        mpz(7)
        """
        cdef FLINT_Integer c
        c = self.thisptr.inhomo
        return FLINT_Integer_to_Python(c)

    def coefficient(self, v):   
        """
        Return the coefficient of the variable ``v``.

        INPUT:

        - ``v`` -- a :class:`Variable`.

        OUTPUT:

        An Integer. 

        Examples::

        >>> from pplite import Variable, Affine_Expression
        >>> e = Variable(2) - 3*Variable(4)
        >>> a = Affine_Expression(e, 7)
        >>> a.coefficient(Variable(2))
        mpz(1)
        """
        return self.linear_form().coefficient(v)

    def __repr__(self):
        r"""
        Return a string representation of the linear expression.

        OUTPUT:

        A string.

        Examples::

        >>> from pplite import Linear_Expression, Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> x
        x0
        >>> x-x
        0
        >>> 2*x
        2*x0
        """
        s = ''
        first = True
        for i in range(self.space_dimension()):
            x = Variable(i)
            coeff = self.coefficient(x)
            if coeff == 0:
                continue
            if first and coeff == 1:
                s += '%r' % x
                first = False
            elif first and coeff == -1:
                s += '-%r' % x
                first = False
            elif first and coeff != 1:
                s += '%d*%r' % (coeff, x)
                first = False
            elif coeff == 1:
                s += '+%r' % x
            elif coeff == -1:
                s += '-%r' % x
            else:
                s += '%+d*%r' % (coeff, x)
        inhomog = self.inhomogeneous_term()
        if inhomog != 0:
            if first:
                s += '%d' % inhomog
                first = False
            else:
                s += '%+d' % inhomog
        if first:
            s = '0'
        return s

    def all_terms_are_zero(self):
        """
        Test if ``self`` is a constant linear expression.

        OUTPUT:

        boolean
        """
        return self.thisptr.is_zero()

    def is_equal_to(self, Affine_Expression other):
        """
        Test equality with another affine expression.

        OUTPUT: 

        boolean
        """
        if self.inhomogeneous_term() == other.inhomogeneous_term() and self.linear_form().is_equal_to(other.linear_form()):
            return True
        return False

    def m_swap(self, Affine_Expression y):
        self.thisptr[0].m_swap(y.thisptr[0])
        return self

    def normalize(self):
        """
        Normalize the Affine_expression.

        """
        self.thisptr[0].normalize()

    def sign_normalize(self):
        self.thisptr[0].sign_normalize()

    def __add__(self, other):
        r"""
        Add ``self`` and ``other``.

        INPUT:

        - ``self``, ``other`` -- anything that can be used to
          construct a :class:`Affine_Expression`. One of them, not
          necessarily ``self``, is guaranteed to be a
          :class:``Affine_Expression``, otherwise Python would not
          have called this method.

        OUTPUT:

        The sum as a :class:`Affine_Expression`

        Examples::

        >>> from pplite import Affine_Expression, Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> a = x + y + y + y + 1; a
        x0+3*x1+1
        >>> a.inhomogeneous_term()
        mpz(1)
        >>> a + a
        2*x0+6*x1+2
        >>> 12 + a
        x0+3*x1+13
        >>> a + Variable(2)
        x0+3*x1+x2+1
        >>> a.linear_form()
        x0+3*x1
        >>> a.linear_form()+ a.inhomogeneous_term()
        x0+3*x1+1
        """
        cdef Affine_Expr* lhs
        if isinstance(self, Affine_Expression):
            lhs = (<Affine_Expression> self).thisptr
        else:
            lhs_expr = Affine_Expression(self)
            lhs = (<Affine_Expression> lhs_expr).thisptr
        result_expr = Affine_Expression()
        cdef Var* vv
        cdef Affine_Expr* temp
        if isinstance(other, Variable):
            other_var = <Variable> other
            vv = other_var.thisptr
            temp = new Affine_Expr(self.thisptr[0])
            self.thisptr[0] += vv[0]  # There is an oddity where adding affine expression and variable doesn't work.
            # This means to add via PPlite add methods, we need to use += operator (it is what works okay!). This modifies self's data. 
            # To ensure that the original affine expression's data is unmodified, temporary store data to save it.
            # This has to do with cython and overloaded operators. If this is issue is ever fixed, this could be revisited.  
            result_expr.thisptr = new Affine_Expr(self.thisptr[0]) 
            self.thisptr = temp
            return result_expr
        cdef Linear_Expr* e1
        if isinstance(other, Linear_Expression):
            other_lin = <Linear_Expression> other
            e1 = other_lin.thisptr
            self.thisptr[0] += e1[0]
            result_expr.thisptr = new Affine_Expr(self.thisptr[0])
            return result_expr
        # TODO: Make adding int explicit. 
        if isinstance(other, Affine_Expression):
            rhs = (<Affine_Expression> other).thisptr
        else:
            rhs_expr = Affine_Expression(other)
            rhs = (<Affine_Expression> rhs_expr).thisptr
        cdef Affine_Expr result
        result = lhs[0] + rhs[0]
        result_expr = Affine_Expression()
        result_expr.thisptr[0] = result
        return result_expr

    def __radd__(self, other):
        return self + other

    def __sub__(self, other):
        r"""
        Subtract ``self`` and ``other``.

        INPUT:

        - ``self``, ``other`` -- anything that can be used to
          construct a :class:`Affine_Expression`. One of them, not
          necessarily ``self``, is guaranteed to be a
          :class:``Affine_Expression``, otherwise Python would not
          have called this method.

        OUTPUT:

        The sum as a :class:`Affine_Expression`

        Examples:

        >>> from pplite import Affine_Expression, Linear_Expression, Variable
        >>> x = Variable(0)
        >>> y = Variable(1)
        >>> e = x + y + y + y + 1; e
        x0+3*x1+1
        >>> e - e
        0
        >>> isinstance(e, Affine_Expression)
        True
        >>> 12 - e
        -x0-3*x1+11
        >>> e - 3 * Variable(3) - 3 * y
        x0-3*x3+1
        """
        cdef Affine_Expr* lhs
        if isinstance(self, Affine_Expression):
            lhs = (<Affine_Expression> self).thisptr
        else:
            lhs_expr = Affine_Expression(self)
            lhs = (<Affine_Expression> lhs_expr).thisptr
        result_expr = Affine_Expression()
        cdef Var* vv
        if isinstance(other, Variable):
            other_var = <Variable> other
            vv = other_var.thisptr
            self.thisptr[0] -= vv[0]
            result_expr.thisptr = new Affine_Expr(self.thisptr[0])
            return result_expr
        cdef Linear_Expr* e1
        if isinstance(other, Linear_Expression):
            other_lin = <Linear_Expression> other
            e1 = other_lin.thisptr
            self.thisptr[0] -= e1[0]
            result_expr.thisptr = new Affine_Expr(self.thisptr[0])
            return result_expr
        # TODO: Make adding int explicit.  
        if isinstance(other, Affine_Expression):
            rhs = (<Affine_Expression> other).thisptr
        else:
            rhs_expr = Affine_Expression(other)
            rhs = (<Affine_Expression> rhs_expr).thisptr
        cdef Affine_Expr result
        result = lhs[0] - rhs[0]
        result_expr = Affine_Expression()
        result_expr.thisptr[0] = result
        return result_expr

    def __rsub__(self, other):
        cdef Affine_Expr* lhs
        if isinstance(self, Affine_Expression):
            lhs = (<Affine_Expression> self).thisptr
        else:
            lhs_expr = Affine_Expression(self)
            lhs = (<Affine_Expression> lhs_expr).thisptr
        result_expr = Affine_Expression()
        cdef Var* vv
        if isinstance(other, Variable):
            other_var = <Variable> other
            vv = other_var.thisptr
            result_expr.thisptr = new Affine_Expr(vv[0] - self.thisptr[0]) 
            return result_expr
        # TODO: Make adding int explicit.
        if isinstance(other, Affine_Expression):
            rhs = (<Affine_Expression> other).thisptr
        else:
            rhs_expr = Affine_Expression(other)
            rhs = (<Affine_Expression> rhs_expr).thisptr
        cdef Affine_Expr result
        result = rhs[0] - lhs[0]
        result_expr = Affine_Expression()
        result_expr.thisptr[0] = result
        return result_expr

    def __mul__(self, other):
        cdef Affine_Expr* a
        if isinstance(self, Linear_Expression):
            a = (<Affine_Expression> self).thisptr
            c = other
        else:
            # NOTE: this code path will only be executed when compiled with cython < 3.0.0
            a = (<Affine_Expression> other).thisptr
            c = self
        cdef FLINT_Integer cc = Python_int_to_FLINT_Integer(c)
        cdef Affine_Expression result = Linear_Expression()
        result.thisptr[0] = a[0] * cc
        return result

    def __rmul__(self, other):
        cdef Affine_Expr* a
        if isinstance(self, Linear_Expression):
            a = (<Affine_Expression> self).thisptr
            c = other
        else:
            # NOTE: this code path will only be executed when compiled with cython < 3.0.0
            a = (<Affine_Expression> other).thisptr
            c = self
        cdef FLINT_Integer cc = Python_int_to_FLINT_Integer(c)
        cdef Affine_Expression result = Linear_Expression()
        result.thisptr[0] = a[0] * cc
        return result

    def __neg__(self):
        neg_assign(self.thisptr[0])
        return self

    def __pos__(self):
        return self

    def __richcmp__(self, other, op):
        """
        Construct :class:`Constraint`s

        Examples:

        >>> from pplite import Variable
        """
        return _make_Constraint_from_richcmp(self, other, op)


# distutils: extra_compile_args = -std=c++11

from __future__ import absolute_import

from .pplite_decl cimport *

cdef class Interval(object):
    cdef Itv interval

# cpdef object Interval():

# cdef enum Spec_Elem "":
#     EMPTY
#     UNIVERSE

# cdef enum Kind "Itv::Kind":
#     UNIVERSE
#     L_BOUNDED
#     U_BOUNDED
#     LU_BOUNDED
#     EMPTY



# distutils: extra_compile_args = -std=c++11

from __future__ import absolute_import

from .pplite_decl cimport *

from .constraint cimport *

cdef class PPliteGenerator(object):
    cdef Gen *thisptr

cdef enum GenType:
    LINE
    RAY
    POINT
    CLOSURE_POINT